import collections
import os.path
import pickle
import queue
import warnings

import torch.cuda
import torch.multiprocessing as mlp
import flgo
import flgo.algorithm.fedavg as fedavg
import zmq
import time
import flgo.utils.fmodule
import threading
import numpy as np

def default_start_condition(server):
    return server.num_clients>=6

class Server(fedavg.Server):
    def __init__(self, option={}):
        self.test_data = None
        self.val_data = None
        self.train_data = None
        self.model = None
        # basic configuration
        self.task = option['task']
        self.eval_interval = option['eval_interval']
        self.num_parallels = option['num_parallels']
        # server calculator
        self.device = self.gv.apply_for_device() if not option['server_with_cpu'] else torch.device('cpu')
        self.calculator = self.TaskCalculator(self.device, optimizer_name=option['optimizer'])
        # hyper-parameters during training process
        self.num_rounds = option['num_rounds']
        self.num_steps = option['num_steps']
        self.num_epochs = option['num_epochs']
        self.proportion = option['proportion']
        self.decay_rate = option['learning_rate_decay']
        self.lr_scheduler_type = option['lr_scheduler']
        self.lr = option['learning_rate']
        self.learning_rate = option['learning_rate']
        self.sample_option = option['sample']
        self.aggregation_option = option['aggregate']
        # systemic option
        self.tolerance_for_latency = 999999
        # algorithm-dependent parameters
        self.algo_para = {}
        self.current_round = 1
        # all options
        self.option = option
        self.id = -1
        self._data_names = []
        self._exit = False
        self._avalability_timeout = 30
        self._communication_timeout = 1e10
        self.start_condition = default_start_condition

    def register_start_condition(self, f):
        assert callable(f)
        self.start_condition = f

    def set_availability_timeout(self, t:float):
        r"""
        Set the timeout of being unavailable for each client.

        Args:
            t (float): the time (secs)
        """
        self._avalability_timeout = t

    def set_communication_timeout(self, t:float):
        r"""
        Set the timeout of waiting for clients' responses

        Args:
            t (float): the time (secs)
        """
        self._communication_timeout = t

    def is_exit(self):
        r"""
        Return True when the server is going to close
        """
        self._lock_exit.acquire()
        res = self._exit
        self._lock_exit.release()
        return res

    def do_exit(self):
        r"""
        Set the flag of exit to be True
        """
        self._lock_exit.acquire()
        self._exit = True
        self._lock_exit.release()
        self.logger.info("Server was closed")
        return

    def add_buffer(self, x:dict):
        r"""
        Push x into network buffer (i.e. a queue)

        Args:
            x (dict): received packages from clients
        """
        self._lock_buffer.acquire()
        self._buffer.put(x)
        self._lock_buffer.release()
        return

    def clear_buffer(self):
        r"""
        Pop all elements in the buffer

        Return:
            res (dict): a dict contains pairs of (client_name, package) in buffer
        """
        self._lock_buffer.acquire()
        res = {}
        while not self._buffer.empty():
            d = self._buffer.get_nowait()
            res[d['name']] = d['package']
        self._lock_buffer.release()
        return res

    def size_buffer(self):
        r"""
        Return the number of current elements in buffer

        Return:
            buffer_size (int): the number of elements
        """
        self._lock_buffer.acquire()
        buffer_size = len(self._buffer)
        self._lock_buffer.release()
        return buffer_size

    def register(self):
        self.logger.info("Waiting for registrations...")
        while True:
            time.sleep(1)
            if self.if_start():
                self.logger.info("Start training...")
                break
        return

    def if_start(self):
        return self.start_condition(self)

    def register_handler(self, worker_id, client_id, received_pkg):
        valid_keys = ['num_steps', 'learning_rate', 'batch_size', 'momentum', 'weight_decay', 'num_epochs', 'optimizer']
        self._set_alive(received_pkg["name"], time.time())
        if received_pkg["name"] not in self.clients.keys():
            self.add_client(received_pkg["name"])
            l = len(self.clients)
            self.logger.info("%s joined in the federation. The number of clients is %i" % (received_pkg['name'], l))

            d = {"client_idx": l, 'port_send': self.port_send, 'port_recv': self.port_recv, 'port_alive':self.port_alive,
                 '__option__': {k: self.option[k] for k in valid_keys},'algo_para':self.algo_para}
            self.registrar.send_multipart([worker_id, client_id, pickle.dumps(d, pickle.DEFAULT_PROTOCOL)])
        else:
            self.logger.info("%s rebuilt the connection." % received_pkg['name'])
            self.registrar.send_pyobj({"client_idx": len(self.clients), 'port_send': self.port_send, 'port_recv': self.port_recv, 'port_alive':self.port_alive, '__option__': {k: self.option[k] for k in valid_keys}, 'algo_para':self.algo_para})

    def task_pusher_handler(self, worker_id, client_id):
        zipped_task = self._get_zipped_task()
        if zipped_task is None:
            self._read_zipped_task()
            zipped_task = self._get_zipped_task()
        self.task_pusher.send_multipart([worker_id, client_id, zipped_task])
        return

    def _listen(self):
        while not self.is_exit():
            events = dict(self._poller.poll(10000))
            if self.task_pusher in events and events[self.task_pusher]==zmq.POLLIN:
                worker_id, client_id, name, request = self.task_pusher.recv_multipart()
                if request==b'pull task':
                    try:
                        self.logger.info("Receive task pull request from {}".format(name))
                        t = threading.Thread(target=self.task_pusher_handler, args=(worker_id, client_id))
                        t.start()
                    except:
                        self.logger.info("Failed to handle task for %s" % client_id)
            if self.registrar in events and events[self.registrar]==zmq.POLLIN:
                worker_id = self.registrar.recv()
                client_id = self.registrar.recv()
                received_pkg = self.registrar.recv_pyobj()
                t = threading.Thread(target=self.register_handler, args=(worker_id, client_id, received_pkg))
                t.start()
            if self.receiver in events and events[self.receiver]==zmq.POLLIN:
                name = self.receiver.recv_string()
                package_msg = self.receiver.recv()
                package_size = len(package_msg) / 1024.0 / 1024.0
                d = self.receiver._deserialize(package_msg, pickle.loads)
                d['__size__'] = package_size
                if '__mtype__' in d and d['__mtype__'] == "close":
                    self.logger.info("{} was successfully closed.".format(name))
                else:
                    self.add_buffer({'name': name, 'package': d})
                    self.logger.info("Server Received package of size {}MB from {} at round {}".format(package_size, name, self.current_round))
            if self.alive_detector in events and events[self.alive_detector]==zmq.POLLIN:
                name, _ = self.alive_detector.recv_multipart()
                t = time.time()
                self.logger.debug(f"Client {name.decode('utf-8')} is alive at {time.strftime('%Y-%m-%d %H:%M:%S')}")
                self._set_alive(name.decode('utf-8'), t)
                self.alive_detector.send(b"")
    @property
    def clients(self):
        self._lock_registration.acquire()
        res = self._clients
        self._lock_registration.release()
        return res

    @property
    def num_clients(self):
        return len(self.clients)

    def add_client(self, name):
        self._lock_registration.acquire()
        i = len(self._clients)
        self._clients[i] = name
        self._lock_registration.release()
        return

    def _set_alive(self, client_name:str, timestamp):
        self._lock_alive.acquire()
        self._alive_state[client_name] = timestamp
        self._lock_alive.release()

    @property
    def available_clients(self):
        crt_timestamp = time.time()
        self._lock_alive.acquire()
        avl_clients = [name for name in self._alive_state if crt_timestamp-self._alive_state[name]<=self._avalability_timeout]
        res = {k:v for k,v in self.clients.items() if v in avl_clients}
        self._lock_alive.release()
        return res

    def pack(self, client_id, mtype=0, *args, **kwargs):
        if mtype=='close':
            return {}
        else:
            return {'model': self.model}

    def _read_zipped_task(self, with_bmk=True, ignore_names=[]):
        task_path = self.option['task']
        task_name = os.path.basename(task_path)
        task_dir = os.path.dirname(os.path.abspath(task_path))
        task_zip = task_name + '.zip'
        if not os.path.exists(task_zip):
            flgo.zip_task(task_path, target_path=task_dir, with_bmk=with_bmk, ignore_names=ignore_names)
        if not hasattr(self, '_zipped_task'): self._zipped_task = []
        CHUNK_SIZE = 1024
        with open(os.path.join(task_dir, task_zip), 'rb') as inf:
            while True:
                chunk = inf.read(CHUNK_SIZE)
                self._zipped_task.append(chunk)
                if not chunk:
                    break

    def _get_zipped_task(self):
        if not hasattr(self, '_zipped_task'): return None
        return b"".join(self._zipped_task)

    def run(self, ip:str='*', port:str='5555', port_task:str='', protocol:str='tcp', no_zip_flags:list=['data.json', '__pycache__', 'checkpoint', 'record', 'log']):
        """
        Start the parameter server process that listens to the public address 'server_ip:server:port' for clients. Each client can connect to this public address to join in training.
        >>> import flgo.algorithm.realbase as realbase
        >>> task = flgo.gen_task(...)
        >>> server_runner = flgo.init(task, realbase, scene='real_hserver')
        >>> server_runner.run(port='5555')
        Args:
            ip (str): ip address
            port (str): public port for client registration, default is 5555
            port_task (str): public port for client pulling task
            protocol (str): the communication protocol, default is TCP
        """
        if 'real' in self.option['scene']: self._read_zipped_task(ignore_names=no_zip_flags)
        self.logger = self.logger(task=self.option['task'], option=self.option, name=self.name+'_'+str(self.logger), level=self.option['log_level'])
        self.logger.register_variable(object=self, server=self)
        self._clients = {}
        self.ip = ip
        self.port = port
        self._lock_registration = threading.Lock()
        self._buffer = queue.Queue()
        self._lock_buffer = threading.Lock()
        self._exit = False
        self._lock_exit = threading.Lock()

        self.context = zmq.Context()
        self.registrar = self.context.socket(zmq.ROUTER)
        self.registrar.bind("%s://%s:%s" % (protocol, ip, port))

        self.port_task = self.get_free_port() if port_task == '' else port_task
        self.task_pusher = self.context.socket(zmq.ROUTER)
        self.task_pusher.bind("%s://%s:%s" % (protocol, ip, self.port_task))
        self.logger.info("Publish Task %s in %s://%s:%s"% (os.path.basename(self.option['task']),protocol, ip, self.port_task))

        self.port_send = self.get_free_port()
        self.sender = self.context.socket(zmq.PUB)
        self.sender.bind("%s://%s:%s" %(protocol, ip, self.port_send))

        self.port_recv = self.get_free_port()
        self.receiver = self.context.socket(zmq.PULL)
        self.receiver.bind("%s://%s:%s"%(protocol, ip, self.port_recv))

        self.port_alive = self.get_free_port()
        self.alive_detector = self.context.socket(zmq.ROUTER)
        self.alive_detector.bind("%s://%s:%s"%(protocol, ip, self.port_alive))
        self._lock_alive = threading.Lock()
        self._alive_state = {}

        self._poller = zmq.Poller()
        self._poller.register(self.task_pusher, zmq.POLLIN)
        self._poller.register(self.registrar, zmq.POLLIN)
        self._poller.register(self.receiver, zmq.POLLIN)
        self._poller.register(self.alive_detector, zmq.POLLIN)
        self._thread_listening = threading.Thread(target=self._listen)
        self._thread_listening.start()

        self.register()

        self.current_round = 1
        self.logger.time_start('Total Time Cost')
        if not self._load_checkpoint() and self.eval_interval>0:
            # evaluating initial model performance
            self.logger.info("--------------Initial Evaluation--------------")
            self.logger.time_start('Eval Time Cost')
            self.logger.log_once()
            self.logger.time_end('Eval Time Cost')
        while self.current_round<=self.num_rounds:
            updated = self.iterate()
            if updated is True or updated is None:
                self.logger.info("--------------Round {}--------------".format(self.current_round))
                if self.logger.check_if_log(self.current_round, self.eval_interval):
                    self.logger.time_start('Eval Time Cost')
                    self.logger.log_once()
                    self.logger.time_end('Eval Time Cost')
                    self._save_checkpoint()
                # check if early stopping
                if self.logger.early_stop(): break
                self.current_round += 1
        self.logger.info("=================End==================")
        self.logger.time_end('Total Time Cost')
        # save results as .json file
        self.logger.save_output_as_json()
        self.do_exit()
        self.communicate([_ for _ in range(len(self.clients))], mtype='close')
        exit(0)

    def aggregate(self, models: list, *args, **kwargs):
        if len(models)==0: return self.model
        return flgo.utils.fmodule._model_average(models).to(self.device)

    def unpack(self, pkgs:dict):
        if len(pkgs)==0: return collections.defaultdict(list)
        keys = list(list(pkgs.values())[0].keys())
        res = {}
        for k in keys:
            res[k] = []
            for cip in pkgs:
                v = pkgs[cip][k]
                res[k].append(v)
        return res

    def communicate_with(self, target_id, package={}):
        self.sender.send_string(target_id, zmq.SNDMORE)
        self.sender.send_pyobj(package)

    def communicate(self, selected_clients, mtype=0, asynchronous=False, only_available=True):
        avl_clients = self.available_clients
        avl_selected_clients = []
        uavl_selected_clients = []
        for i in selected_clients:
            if i in avl_clients.keys():
                avl_selected_clients.append(i)
            else:
                uavl_selected_clients.append(i)
        if only_available:
            self.selected_clients = avl_selected_clients
            for i in uavl_selected_clients:
                self.logger.info(f"Selected client {self.clients[i]} is dropped since it is currently not available.")
        else:
            for i in uavl_selected_clients:
                warnings.warn(f"Selected client {self.clients[i]} is dropped since it is currently not available.")
        selected_clients = [self.clients[i] for i in self.selected_clients]
        self.model.to('cpu')
        for i, name in enumerate(selected_clients):
            package = self.pack(i, mtype=mtype)
            package['__mtype__'] = mtype
            package['__round__'] = self.current_round
            package['name'] = name
            self.communicate_with(name, package)
        buffer = {}
        start_timestamp = time.time()
        while not self.is_exit():
            new_comings = self.clear_buffer()
            buffer.update(new_comings)
            crt_cost = time.time() - start_timestamp
            if asynchronous or all([(name in buffer) for name in selected_clients]) or crt_cost>self._communication_timeout:
                if crt_cost>self._communication_timeout:
                    timeout_clients = [name for name in selected_clients if name not in buffer.keys()]
                    for name in timeout_clients:
                        self.logger.info(f"Failed to receive packages from Client {name} due to timeout {self._communication_timeout}s. Using 'set_communication_timeout' method to set timeout can allow a longer waiting period.")
                break
            time.sleep(0.1)
        return self.unpack(buffer)

    def global_test(self, model=None, flag: str = 'val'):
        all_metrics = self.communicate([_ for _ in range(len(self.clients))], mtype='%s_metric'%flag)
        return all_metrics

    def get_free_port(self):
        import socket
        sock = socket.socket()
        sock.bind(('', 0))
        ip, port = sock.getsockname()
        sock.close()
        return port

    def init_algo_para(self, algo_para: dict):
        self.algo_para = algo_para
        if len(self.algo_para) == 0: return
        # initialize algorithm-dependent hyperparameters from the input options
        if self.option['algo_para'] is not None:
            # assert len(self.algo_para) == len(self.option['algo_para'])
            keys = list(self.algo_para.keys())
            for i, pv in enumerate(self.option['algo_para']):
                if i == len(self.option['algo_para']): break
                para_name = keys[i]
                try:
                    self.algo_para[para_name] = type(self.algo_para[para_name])(pv)
                except:
                    self.algo_para[para_name] = pv
        for para_name, value in self.algo_para.items():
            self.__setattr__(para_name, value)

    def save_checkpoint(self):
        cpt = {
            'round': self.current_round,
            'model_state_dict': self.model.state_dict(),
            'early_stop_option': {
                '_es_best_score': self.logger._es_best_score,
                '_es_best_round': self.logger._es_best_round,
                '_es_patience': self.logger._es_patience,
            },
            'output': self.logger.output,
        }
        return cpt

    def load_checkpoint(self, cpt):
        md = cpt.get('model_state_dict', None)
        round = cpt.get('round', None)
        output = cpt.get('output', None)
        early_stop_option = cpt.get('early_stop_option', None)
        time = cpt.get('time', None)
        learning_rate = cpt.get('learning_rate', None)
        if md is not None: self.model.load_state_dict(md)
        if round is not None: self.current_round = round + 1
        if output is not None: self.gv.logger.output = output
        if time is not None: self.gv.clock.set_time(time)
        if learning_rate is not None: self.learning_rate = learning_rate
        if early_stop_option is not None:
            self.gv.logger._es_best_score = early_stop_option['_es_best_score']
            self.gv.logger._es_best_round = early_stop_option['_es_best_round']
            self.gv.logger._es_patience = early_stop_option['_es_patience']

class Client(fedavg.Client):
    def __init__(self, option={}):
        super(Client, self).__init__(option)
        self.actions = {
            '0': self.reply,
        }
        self.timeout_register = 5
        self.heartbeat_interval = 30

    def val_metric(self, package):
        model = package['model']
        metrics = self.test(model, 'val')
        return metrics

    def test_metric(self, package):
        model = package['model']
        metrics = self.test(model, 'test')
        return metrics

    def train_metric(self, package):
        model = package['model']
        metrics = self.test(model, 'train')
        return metrics

    def set_timeout_register(self, t=5):
        self.timeout_register = t

    def register(self):
        self.logger.info("%s Registering..." % self.name)
        if self.timeout_register>0:
            self.registrar.setsockopt(zmq.RCVTIMEO, self.timeout_register*1000)
        try:
            self.registrar.send_pyobj({"name": self.name})
            reply = self.registrar.recv_pyobj()
        except zmq.error.Again as e:
            self.logger.info("Server temporarily unavailable")
            self.do_exit()
            exit(0)

        if '__option__' in reply: self.set_option(reply['__option__'])
        if 'algo_para' in reply and isinstance(reply['algo_para'], dict):
            self.option['algo_para'] = reply['algo_para']
            for k,v in reply['algo_para'].items(): setattr(self, k, v)
        return reply["port_recv"], reply["port_send"], reply['port_alive']

    def message_handler(self, package, *args, **kwargs):
        mtype = package['__mtype__']
        if package['__mtype__'] == 'close':
            self.sender.send_string(self.name, zmq.SNDMORE)
            self.sender.send_pyobj({'__mtype__': "close"})
            return True
        action = self.default_action if mtype not in self.actions else self.actions[mtype]
        response = action(package)
        assert isinstance(response, dict)
        response['__name__'] = self.name
        if hasattr(self, 'round'): response['__round__'] = self.round
        self.sender.send_string(self.name, zmq.SNDMORE)
        msg = pickle.dumps(response, pickle.DEFAULT_PROTOCOL)
        self.logger.info("{} Sending the package of size {}MB to the server...".format(self.name, len(msg)/1024/1024))
        self.sender.send(msg)
        # self.sender.send_pyobj(response)
        return False

    def is_exit(self):
        self._lock_exit.acquire()
        res = self._exit
        self._lock_exit.release()
        return res

    def do_exit(self):
        self._lock_exit.acquire()
        self._exit = True
        self._lock_exit.release()
        self.logger.info("%s was closed" % self.name)
        return

    def _heart_beat(self):
        while not self.is_exit():
            try:
                time.sleep(self.heartbeat_interval)
                self.heart_beator.send(b"")
            except Exception as e:
                self.logger.info(e)
                continue

    def _listen(self):
        while not self.is_exit():
            events = dict(self._poller.poll(10000))
            if self.heart_beator in events and events[self.heart_beator] == zmq.POLLIN:
                server_is_alive = self.heart_beator.recv()
            if self.receiver in events and events[self.receiver]==zmq.POLLIN:
                name = self.receiver.recv_string()
                assert name==self.name
                package_msg = self.receiver.recv()
                package_size = len(package_msg)
                package = self.receiver._deserialize(package_msg, pickle.loads)
                # package = self.receiver.recv_pyobj()
                assert '__mtype__' in package
                package['__size__'] = package_size/1024/1024 #MB
                if '__round__' in package.keys():
                    self.round = package['__round__']
                    self.logger.info("{} is selected at round {} and has received the package of {}MB".format(self.name, package['__round__'], package['__size__']))
                do_break = self.message_handler(package)
                if do_break: break
        return

    def run(self, server_ip:str='127.0.0.1', server_port: str='5555', protocol:str='tcp', timeout_register=10):
        """
        Start the client process that connects to the public address 'server_ip:server:port' of the parameter server..
        >>> import flgo.algorithm.realbase as realbase
        >>> import flgo
        >>> server_ip = '127.0.0.1'
        >>> server_task_port = ...
        >>> server_register_port = '5555'
        >>> task = ...
        >>> flgo.pull_task_from_("tcp://{}:{}".format(server_ip, server_task_port), task)
        >>> # set local dataset in task/dataset.py before joining in training
        >>> client_runner = flgo.init(task, realbase, scene='real_hclient')
        >>> client_runner.run(server_ip, server_register_port)
        Args:
            server_ip (str): ip address
            server_port (str): public port for client registration, default is 5555
            protocol (str): the communication protocol, default is TCP
            timeout_register (int): the timeout for registeration (seconds), defaulf is 10
        """
        self.logger = self.logger(task=self.option['task'], option=self.option, name=self.name+'_'+str(self.logger), level=self.option['log_level'])
        self.logger.register_variable(object=self, clients = [self])

        self.timeout_register = timeout_register
        self._exit = False
        self._lock_exit = threading.Lock()
        self.actions.update({'val_metric': self.val_metric, 'train_metric': self.train_metric, 'test_metric': self.test_metric,})

        self.context = zmq.Context()
        # Registration Socket
        self.registrar = self.context.socket(zmq.REQ)
        self.registrar.connect("%s://%s:%s"%(protocol, server_ip, server_port))
        port_svr_recv, port_svr_send, port_svr_alive = self.register()
        self.logger.info(f"{self.name} Successfully Registered to {server_ip}:{server_port}")

        self.receiver = self.context.socket(zmq.SUB)
        self.receiver.connect("%s://%s:%s" % (protocol, server_ip, port_svr_send))
        self.receiver.subscribe(self.name)

        self.heart_beator = self.context.socket(zmq.DEALER)
        self.heart_beator.setsockopt(zmq.IDENTITY, self.name.encode('utf-8'))
        self.heart_beator.connect("%s://%s:%s" % (protocol, server_ip, port_svr_alive))

        self.sender = self.context.socket(zmq.PUSH)
        self.sender.connect("%s://%s:%s" % (protocol, server_ip, port_svr_recv))

        self._poller = zmq.Poller()
        self._poller.register(self.receiver, zmq.POLLIN)
        self._poller.register(self.heart_beator, zmq.POLLIN)

        self._thread_heartbeat = threading.Thread(target=self._heart_beat)
        self._thread_heartbeat.start()

        self.logger.info(f"{self.name} Ready For Training...")
        self.logger.info("---------------------------------------------------------------------")
        # threading.Thread(target=self._heart_beat).start()
        self._server_timeout = np.inf
        self._server_alive_timestamp = time.time()
        while True:
            events = dict(self._poller.poll(10000))
            # if self.heart_beator in events and events[self.heart_beator] == zmq.POLLIN:
            #     _ = self.heart_beator.recv()
            #     self._server_alive_timestamp = time.time()
            if self.receiver in events and events[self.receiver]==zmq.POLLIN:
                name = self.receiver.recv_string()
                assert name==self.name
                package_msg = self.receiver.recv()
                package_size = len(package_msg)
                package = self.receiver._deserialize(package_msg, pickle.loads)
                # package = self.receiver.recv_pyobj()
                assert '__mtype__' in package
                if package['__mtype__']=='close':
                    self.sender.send_string(self.name, zmq.SNDMORE)
                    self.sender.send_pyobj({'__mtype__':"close"})
                    break
                package['__size__'] = package_size/1024/1024 #MB
                if '__round__' in package.keys():
                    self.round = package['__round__']
                    self.logger.info("{} is selected at round {} and has received the package of {}MB".format(self.name, package['__round__'], package['__size__']))
                self.message_handler(package)
            if time.time() - self._server_alive_timestamp>=self._server_timeout:
                self.logger.info("Lose connection to the server")
                self.do_exit()
                break
        torch.cuda.empty_cache()
        self.do_exit()
        exit(0)

    def set_option(self, option:dict={}):
        valid_keys = ['num_steps', 'learning_rate', 'batch_size', 'momentum', 'weight_decay', 'num_epochs', 'optimizer']
        types = [int, float, float, float, float, int, str]
        self.option.update(option)
        for k,t in zip(valid_keys, types):
            if k in option:
                try:
                    setattr(self, k, t(option[k]))
                except:
                    self.logger.info("Failed to set hyper-parameter {}={}".format(k, option[k]))
                    continue
        # correct hyper-parameters
        if hasattr(self, 'train_data') and self.train_data is not None:
            import math
            self.datavol = len(self.train_data)
            if hasattr(self, 'batch_size'):
                # reset batch_size
                if self.batch_size < 0:
                    self.batch_size = self.datavol
                elif self.batch_size >= 1:
                    self.batch_size = int(self.batch_size)
                else:
                    self.batch_size = int(self.datavol * self.batch_size)
            # reset num_steps
            if hasattr(self, 'num_steps') and hasattr(self, 'num_epochs'):
                if self.num_steps > 0:
                    self.num_epochs = 1.0 * self.num_steps / (math.ceil(self.datavol / self.batch_size))
                else:
                    self.num_steps = self.num_epochs * math.ceil(self.datavol / self.batch_size)
        return

    def set_data(self, data, flag:str='train') -> None:
        r"""
        Set self's attibute 'xxx_data' to be data where xxx is the flag. For example,
        after calling self.set_data([1,2,3], 'test'), self.test_data will be [1,2,3].
        Particularly, If the flag is 'train', the batchsize and the num_steps will be
        reset.

        Args:
            data: anything
            flag (str): the name of the data
        """
        setattr(self, flag + '_data', data)
        if flag not in self._data_names:
            self._data_names.append(flag)
        if flag == 'train':
            if data is None:
                warnings.warn("Local train data is None")
                return
            import math
            self.datavol = len(data)
            if hasattr(self, 'batch_size'):
                # reset batch_size
                if self.batch_size < 0:
                    self.batch_size = len(self.get_data(flag))
                elif self.batch_size >= 1:
                    self.batch_size = int(self.batch_size)
                else:
                    self.batch_size = int(self.datavol * self.batch_size)
            # reset num_steps
            if hasattr(self, 'num_steps') and hasattr(self, 'num_epochs'):
                if self.num_steps > 0:
                    self.num_epochs = 1.0 * self.num_steps / (math.ceil(self.datavol / self.batch_size))
                else:
                    self.num_steps = self.num_epochs * math.ceil(self.datavol / self.batch_size)

if __name__=='__main__':
    class algo:
        Server = Server
        Client = Client
    mlp.set_start_method('spawn', force=True)
    mlp.set_sharing_strategy('file_system')
    import flgo.benchmark.mnist_classification as mnist
    import flgo.benchmark.partition as fbp
    flgo.gen_task_by_(mnist, fbp.IIDPartitioner(num_clients=10), 'my_task')
    runner = flgo.init('my_task', algo, option={'proportion':0.2, 'gpu':[0], 'server_with_cpu':True, 'num_rounds':10, 'num_steps':1, 'log_file':True, 'log_level':'DEBUG'}, scene='parallel_horizontal')
    runner.run()


import torch
import torch.nn as nn
import torch.nn.functional as F
from time import time
import numpy as np

def timeit(tag, t):
    print("{}: {}s".format(tag, time() - t))
    return time()

def pc_normalize(pc):
    l = pc.shape[0]
    centroid = np.mean(pc, axis=0)
    pc = pc - centroid
    m = np.max(np.sqrt(np.sum(pc**2, axis=1)))
    pc = pc / m
    return pc

def square_distance(src, dst):
    """
    Calculate Euclid distance between each two points.

    src^T * dst = xn * xm + yn * ym + zn * zm；
    sum(src^2, dim=-1) = xn*xn + yn*yn + zn*zn;
    sum(dst^2, dim=-1) = xm*xm + ym*ym + zm*zm;
    dist = (xn - xm)^2 + (yn - ym)^2 + (zn - zm)^2
         = sum(src**2,dim=-1)+sum(dst**2,dim=-1)-2*src^T*dst

    Input:
        src: source points, [B, N, C]
        dst: target points, [B, M, C]
    Output:
        dist: per-point square distance, [B, N, M]
    """
    B, N, _ = src.shape
    _, M, _ = dst.shape
    dist = -2 * torch.matmul(src, dst.permute(0, 2, 1))
    dist += torch.sum(src ** 2, -1).view(B, N, 1)
    dist += torch.sum(dst ** 2, -1).view(B, 1, M)
    return dist


def index_points(points, idx):
    """

    Input:
        points: input points data, [B, N, C]
        idx: sample index data, [B, S]
    Return:
        new_points:, indexed points data, [B, S, C]
    """
    device = points.device
    B = points.shape[0]
    view_shape = list(idx.shape)
    view_shape[1:] = [1] * (len(view_shape) - 1)
    repeat_shape = list(idx.shape)
    repeat_shape[0] = 1
    batch_indices = torch.arange(B, dtype=torch.long).to(device).view(view_shape).repeat(repeat_shape)
    new_points = points[batch_indices, idx, :]
    return new_points


def farthest_point_sample(xyz, npoint):
    """
    Input:
        xyz: pointcloud data, [B, N, 3]
        npoint: number of samples
    Return:
        centroids: sampled pointcloud index, [B, npoint]
    """
    device = xyz.device
    B, N, C = xyz.shape
    centroids = torch.zeros(B, npoint, dtype=torch.long).to(device)
    distance = torch.ones(B, N).to(device) * 1e10
    farthest = torch.randint(0, N, (B,), dtype=torch.long).to(device)
    batch_indices = torch.arange(B, dtype=torch.long).to(device)
    for i in range(npoint):
        centroids[:, i] = farthest
        centroid = xyz[batch_indices, farthest, :].view(B, 1, 3)
        dist = torch.sum((xyz - centroid) ** 2, -1)
        mask = dist < distance
        distance[mask] = dist[mask]
        farthest = torch.max(distance, -1)[1]
    return centroids


def query_ball_point(radius, nsample, xyz, new_xyz):
    """
    Input:
        radius: local region radius
        nsample: max sample number in local region
        xyz: all points, [B, N, 3]
        new_xyz: query points, [B, S, 3]
    Return:
        group_idx: grouped points index, [B, S, nsample]
    """
    device = xyz.device
    B, N, C = xyz.shape
    _, S, _ = new_xyz.shape
    group_idx = torch.arange(N, dtype=torch.long).to(device).view(1, 1, N).repeat([B, S, 1])
    sqrdists = square_distance(new_xyz, xyz)
    group_idx[sqrdists > radius ** 2] = N
    group_idx = group_idx.sort(dim=-1)[0][:, :, :nsample]
    group_first = group_idx[:, :, 0].view(B, S, 1).repeat([1, 1, nsample])
    mask = group_idx == N
    group_idx[mask] = group_first[mask]
    return group_idx


def sample_and_group(npoint, radius, nsample, xyz, points, returnfps=False):
    """
    Input:
        npoint:
        radius:
        nsample:
        xyz: input points position data, [B, N, 3]
        points: input points data, [B, N, D]
    Return:
        new_xyz: sampled points position data, [B, npoint, nsample, 3]
        new_points: sampled points data, [B, npoint, nsample, 3+D]
    """
    B, N, C = xyz.shape
    S = npoint
    fps_idx = farthest_point_sample(xyz, npoint) # [B, npoint, C]
    new_xyz = index_points(xyz, fps_idx)
    idx = query_ball_point(radius, nsample, xyz, new_xyz)
    grouped_xyz = index_points(xyz, idx) # [B, npoint, nsample, C]
    grouped_xyz_norm = grouped_xyz - new_xyz.view(B, S, 1, C)

    if points is not None:
        grouped_points = index_points(points, idx)
        new_points = torch.cat([grouped_xyz_norm, grouped_points], dim=-1) # [B, npoint, nsample, C+D]
    else:
        new_points = grouped_xyz_norm
    if returnfps:
        return new_xyz, new_points, grouped_xyz, fps_idx
    else:
        return new_xyz, new_points


def sample_and_group_all(xyz, points):
    """
    Input:
        xyz: input points position data, [B, N, 3]
        points: input points data, [B, N, D]
    Return:
        new_xyz: sampled points position data, [B, 1, 3]
        new_points: sampled points data, [B, 1, N, 3+D]
    """
    device = xyz.device
    B, N, C = xyz.shape
    new_xyz = torch.zeros(B, 1, C).to(device)
    grouped_xyz = xyz.view(B, 1, N, C)
    if points is not None:
        new_points = torch.cat([grouped_xyz, points.view(B, 1, N, -1)], dim=-1)
    else:
        new_points = grouped_xyz
    return new_xyz, new_points


class PointNetSetAbstraction(nn.Module):
    def __init__(self, npoint, radius, nsample, in_channel, mlp, group_all):
        super(PointNetSetAbstraction, self).__init__()
        self.npoint = npoint
        self.radius = radius
        self.nsample = nsample
        self.mlp_convs = nn.ModuleList()
        self.mlp_bns = nn.ModuleList()
        last_channel = in_channel
        for out_channel in mlp:
            self.mlp_convs.append(nn.Conv2d(last_channel, out_channel, 1))
            self.mlp_bns.append(nn.BatchNorm2d(out_channel))
            last_channel = out_channel
        self.group_all = group_all

    def forward(self, xyz, points):
        """
        Input:
            xyz: input points position data, [B, C, N]
            points: input points data, [B, D, N]
        Return:
            new_xyz: sampled points position data, [B, C, S]
            new_points_concat: sample points feature data, [B, D', S]
        """
        xyz = xyz.permute(0, 2, 1)
        if points is not None:
            points = points.permute(0, 2, 1)

        if self.group_all:
            new_xyz, new_points = sample_and_group_all(xyz, points)
        else:
            new_xyz, new_points = sample_and_group(self.npoint, self.radius, self.nsample, xyz, points)
        # new_xyz: sampled points position data, [B, npoint, C]
        # new_points: sampled points data, [B, npoint, nsample, C+D]
        new_points = new_points.permute(0, 3, 2, 1) # [B, C+D, nsample,npoint]
        for i, conv in enumerate(self.mlp_convs):
            bn = self.mlp_bns[i]
            new_points =  F.relu(bn(conv(new_points)))

        new_points = torch.max(new_points, 2)[0]
        new_xyz = new_xyz.permute(0, 2, 1)
        return new_xyz, new_points


class PointNetSetAbstractionMsg(nn.Module):
    def __init__(self, npoint, radius_list, nsample_list, in_channel, mlp_list):
        super(PointNetSetAbstractionMsg, self).__init__()
        self.npoint = npoint
        self.radius_list = radius_list
        self.nsample_list = nsample_list
        self.conv_blocks = nn.ModuleList()
        self.bn_blocks = nn.ModuleList()
        for i in range(len(mlp_list)):
            convs = nn.ModuleList()
            bns = nn.ModuleList()
            last_channel = in_channel + 3
            for out_channel in mlp_list[i]:
                convs.append(nn.Conv2d(last_channel, out_channel, 1))
                bns.append(nn.BatchNorm2d(out_channel))
                last_channel = out_channel
            self.conv_blocks.append(convs)
            self.bn_blocks.append(bns)

    def forward(self, xyz, points):
        """
        Input:
            xyz: input points position data, [B, C, N]
            points: input points data, [B, D, N]
        Return:
            new_xyz: sampled points position data, [B, C, S]
            new_points_concat: sample points feature data, [B, D', S]
        """
        xyz = xyz.permute(0, 2, 1)
        if points is not None:
            points = points.permute(0, 2, 1)

        B, N, C = xyz.shape
        S = self.npoint
        new_xyz = index_points(xyz, farthest_point_sample(xyz, S))
        new_points_list = []
        for i, radius in enumerate(self.radius_list):
            K = self.nsample_list[i]
            group_idx = query_ball_point(radius, K, xyz, new_xyz)
            grouped_xyz = index_points(xyz, group_idx)
            grouped_xyz -= new_xyz.view(B, S, 1, C)
            if points is not None:
                grouped_points = index_points(points, group_idx)
                grouped_points = torch.cat([grouped_points, grouped_xyz], dim=-1)
            else:
                grouped_points = grouped_xyz

            grouped_points = grouped_points.permute(0, 3, 2, 1)  # [B, D, K, S]
            for j in range(len(self.conv_blocks[i])):
                conv = self.conv_blocks[i][j]
                bn = self.bn_blocks[i][j]
                grouped_points =  F.relu(bn(conv(grouped_points)))
            new_points = torch.max(grouped_points, 2)[0]  # [B, D', S]
            new_points_list.append(new_points)

        new_xyz = new_xyz.permute(0, 2, 1)
        new_points_concat = torch.cat(new_points_list, dim=1)
        return new_xyz, new_points_concat


class PointNetFeaturePropagation(nn.Module):
    def __init__(self, in_channel, mlp):
        super(PointNetFeaturePropagation, self).__init__()
        self.mlp_convs = nn.ModuleList()
        self.mlp_bns = nn.ModuleList()
        last_channel = in_channel
        for out_channel in mlp:
            self.mlp_convs.append(nn.Conv1d(last_channel, out_channel, 1))
            self.mlp_bns.append(nn.BatchNorm1d(out_channel))
            last_channel = out_channel

    def forward(self, xyz1, xyz2, points1, points2):
        """
        Input:
            xyz1: input points position data, [B, C, N]
            xyz2: sampled input points position data, [B, C, S]
            points1: input points data, [B, D, N]
            points2: input points data, [B, D, S]
        Return:
            new_points: upsampled points data, [B, D', N]
        """
        xyz1 = xyz1.permute(0, 2, 1)
        xyz2 = xyz2.permute(0, 2, 1)

        points2 = points2.permute(0, 2, 1)
        B, N, C = xyz1.shape
        _, S, _ = xyz2.shape

        if S == 1:
            interpolated_points = points2.repeat(1, N, 1)
        else:
            dists = square_distance(xyz1, xyz2)
            dists, idx = dists.sort(dim=-1)
            dists, idx = dists[:, :, :3], idx[:, :, :3]  # [B, N, 3]

            dist_recip = 1.0 / (dists + 1e-8)
            norm = torch.sum(dist_recip, dim=2, keepdim=True)
            weight = dist_recip / norm
            interpolated_points = torch.sum(index_points(points2, idx) * weight.view(B, N, 3, 1), dim=2)

        if points1 is not None:
            points1 = points1.permute(0, 2, 1)
            new_points = torch.cat([points1, interpolated_points], dim=-1)
        else:
            new_points = interpolated_points

        new_points = new_points.permute(0, 2, 1)
        for i, conv in enumerate(self.mlp_convs):
            bn = self.mlp_bns[i]
            new_points = F.relu(bn(conv(new_points)))
        return new_points


"""This is a non-official implementation of 'Federated Learning with Buffered Asynchronous Aggregation' (http://arxiv.org/abs/2106.06639). """
from flgo.algorithm.asyncbase import AsyncServer
from flgo.algorithm.fedbase import BasicClient
import flgo.utils.fmodule as fmodule
import copy

class Server(AsyncServer):
    def initialize(self):
        self.init_algo_para({'buffer_ratio': 0.1, 'eta': 1.0})
        self.buffer = []

    def package_handler(self, received_packages:dict):
        if self.is_package_empty(received_packages): return False
        received_updates = received_packages['model']
        received_client_taus = [u._round for u in received_updates]
        for cdelta, ctau in zip(received_updates, received_client_taus):
            self.buffer.append((cdelta, ctau))
        if len(self.buffer) >= int(self.buffer_ratio * self.num_clients):
            # aggregate and clear updates in buffer
            taus_bf = [b[1] for b in self.buffer]
            updates_bf = [b[0] for b in self.buffer]
            weights_bf = [(1 + self.current_round - ctau) ** (-0.5) for ctau in taus_bf]
            model_delta = fmodule._model_average(updates_bf, weights_bf) / len(self.buffer)
            self.model = self.model + self.eta * model_delta
            # clear buffer
            self.buffer = []
            return True
        return False

class Client(BasicClient):
    def reply(self, svr_pkg):
        model = self.unpack(svr_pkg)
        global_model = copy.deepcopy(model)
        self.train(model)
        update = model-global_model
        update._round = model._round
        cpkg = self.pack(update)
        return cpkg


from flgo.algorithm.fedbase import BasicServer
from flgo.algorithm.fedbase import BasicClient as Client
import numpy as np

class AsyncServer(BasicServer):
    def __init__(self, option={}):
        super(AsyncServer, self).__init__(option)
        self.concurrent_clients = set()
        self.buffered_clients = set()

    def sample(self):
        """
        Sample clients under the limitation of the maximum numder of concurrent clients.
        Returns:
            Selected clients.
        """
        all_clients = self.available_clients if 'available' in self.sample_option else [cid for cid in range(self.num_clients)]
        all_clients = list(set(all_clients).difference(self.buffered_clients))
        clients_per_round = self.clients_per_round - len(self.concurrent_clients)
        if clients_per_round<=0: return []
        clients_per_round = max(min(clients_per_round, len(all_clients)), 1)
        # full sampling with unlimited communication resources of the server
        if 'full' in self.sample_option:
            return all_clients
        # sample clients
        elif 'uniform' in self.sample_option:
            # original sample proposed by fedavg
            selected_clients = list(np.random.choice(all_clients, clients_per_round, replace=False)) if len(all_clients) > 0 else []
        elif 'md' in self.sample_option:
            # the default setting that is introduced by FedProx, where the clients are sampled with the probability in proportion to their local_movielens_recommendation data sizes
            local_data_vols = [self.clients[cid].datavol for cid in all_clients]
            total_data_vol = sum(local_data_vols)
            p = np.array(local_data_vols) / total_data_vol
            selected_clients = list(np.random.choice(all_clients, clients_per_round, replace=True, p=p)) if len(all_clients) > 0 else []
        return selected_clients

    def package_handler(self, received_packages:dict):
        """
        Handle packages received from clients and return whether the global model is updated in this function.

        Args:
            received_packages (dict): a dict consisting of uploaded contents from clients
        Returns:
            is_model_updated (bool): True if the global model is updated in this function.
        """
        if self.is_package_empty(received_packages): return False
        self.model = self.aggregate(received_packages['model'])
        return True

    def is_package_empty(self, received_packages:dict):
        """
        Check whether the package dict is empty

        Returns:
            is_empty (bool): True if the package dict is empty
        """
        return len(received_packages['__cid']) == 0

    def iterate(self):
        """
        The procedure of the server at each moment. Compared to synchronous methods, asynchronous servers perform iterations in a time-level view instead of a round-level view.

        Returns:
            is_model_updated (bool): True if the global model is updated at the current iteration
        """
        self.selected_clients = self.sample()
        self.concurrent_clients.update(set(self.selected_clients))
        if len(self.selected_clients) > 0: self.gv.logger.info('Select clients {} at time {}.'.format(self.selected_clients, self.gv.clock.current_time))
        self.model._round = self.current_round
        received_packages = self.communicate(self.selected_clients, asynchronous=True)
        self.concurrent_clients.difference_update(set(received_packages['__cid']))
        self.buffered_clients.update(set(received_packages['__cid']))
        if len(received_packages['__cid'])>0: self.gv.logger.info('Receive new models from clients {} at time {}'.format(received_packages['__cid'], self.gv.clock.current_time))
        is_model_updated = self.package_handler(received_packages)
        if is_model_updated: self.buffered_clients = set()
        return is_model_updated


import os
import random
import flgo.benchmark
import torch.nn as nn
import torch_geometric
from torch.nn import Linear
import torch.nn.functional as F
from torch_geometric.nn import GCNConv
from torch_geometric.nn import global_mean_pool

path = os.path.join(flgo.benchmark.data_root, 'MUTAG')
all_data = torch_geometric.datasets.TUDataset(path, name='MUTAG', use_node_attr=True)
all_idxs = [i for i in range(len(all_data))]
random.shuffle(all_idxs)
num_samples = len(all_data)
train_idxs = all_idxs[:int(0.9*num_samples)]
test_idxs = all_idxs[int(0.9*num_samples):]
train_data = all_data[train_idxs]
test_data = all_data[test_idxs]


class GCN(nn.Module):
    def __init__(self, hidden_channels=64):
        super(GCN, self).__init__()
        self.conv1 = GCNConv(7, hidden_channels)
        self.conv2 = GCNConv(hidden_channels, hidden_channels)
        self.conv3 = GCNConv(hidden_channels, hidden_channels)
        self.lin = Linear(hidden_channels, 2)

    def forward(self, data):
        x, edge_index, batch = data.x, data.edge_index, data.batch
        # 1. Obtain node embeddings
        x = self.conv1(x, edge_index)
        x = x.relu()
        x = self.conv2(x, edge_index)
        x = x.relu()
        x = self.conv3(x, edge_index)

        # 2. Readout layer
        x = global_mean_pool(x, batch)  # [batch_size, hidden_channels]

        # 3. Apply a final classifier
        x = F.dropout(x, p=0.5, training=self.training)
        x = self.lin(x)
        return x

def get_model():
    return GCN()


import torchvision
import os

transform = torchvision.transforms.Compose([torchvision.transforms.ToTensor(), torchvision.transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010))])
path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'RAW_DATA', 'CIFAR100')
train_data = torchvision.datasets.CIFAR100(root=path, train=True, download=True, transform=transform)
test_data = torchvision.datasets.CIFAR100(root=path, train=False, download=True, transform=transform)


"""
This is a non-official implementation of AFL proposed in 'Agnostic
Federated Learning (http://arxiv.org/abs/1902.00146)'. We refer to the
Github repo fair_flearn ('https://github.com/litian96/fair_flearn') when
implementing this algorithm.
"""

import flgo.utils.fmodule as fmodule
from flgo.algorithm.fedbase import BasicServer, BasicClient
import numpy as np
import copy
import collections


class Server(BasicServer):
    def initialize(self, *args, **kwargs):
        self.init_algo_para({'learning_rate_lambda': 0.01})
        self.dynamic_lambdas = np.ones(self.num_clients) * 1.0 / self.num_clients
        self.learning_rate = self.option['learning_rate']
        self.result_model = copy.deepcopy(self.model)

    def iterate(self):
        # full sampling
        # training
        res = self.communicate([cid for cid in range(self.num_clients)])
        models, train_losses = res['model'], res['loss']
        grads = [(self.model - model) / self.learning_rate for model in models]
        # aggregate grads
        grad = fmodule._model_average(grads, self.dynamic_lambdas)
        self.model = self.model - self.learning_rate * grad
        # update lambdas
        self.dynamic_lambdas = [lmb_i + self.learning_rate_lambda * loss_i for lmb_i, loss_i in
                                zip(self.dynamic_lambdas, train_losses)]
        self.dynamic_lambdas = self.project(self.dynamic_lambdas)
        # record resulting model
        self.result_model = (self.current_round * self.result_model + self.model) / (self.current_round + 1)
        return

    def project(self, p):
        u = sorted(p, reverse=True)
        res = []
        rho = 0
        for i in range(len(p)):
            if (u[i] + (1.0 / (i + 1)) * (1 - np.sum(np.asarray(u)[:i + 1]))) > 0:
                rho = i + 1
        lmbd = (1.0 / rho) * (1 - np.sum(np.asarray(u)[:rho]))
        for i in range(len(p)):
            res.append(max(p[i] + lmbd, 0))
        return res

    def global_test(self, flag='val'):
        """
        Validate accuracies and losses on clients' local_movielens_recommendation datasets
        Args:
            flag: choose train data or valid data to evaluate
        Returns:
            metrics: a dict contains the lists of each metric_value of the clients
        """
        all_metrics = collections.defaultdict(list)
        for c in self.clients:
            client_metrics = c.test(self.result_model, flag)
            for met_name, met_val in client_metrics.items():
                all_metrics[met_name].append(met_val)
        return all_metrics

    def test(self, model=None, flag='test'):
        if model == None: model = self.result_model
        data = self.test_data if flag == 'test' else self.val_data
        if data is None:
            return {}
        else:
            return self.calculator.test(model, self.test_data, batch_size=self.option['test_batch_size'])


class Client(BasicClient):
    def reply(self, svr_pkg):
        model = self.unpack(svr_pkg)
        train_loss = self.test(model, 'train')['loss']
        self.train(model)
        cpkg = self.pack(model, train_loss)
        return cpkg

    def pack(self, model, loss):
        return {
            "model": model,
            "loss": loss,
        }



import torch.nn as nn
from flgo.utils.fmodule import FModule
import torch

class Model(FModule):
    """Creates a RNN model using LSTM layers for Shakespeare language models (next character prediction task).
      This replicates the model structure in the paper:
      Communication-Efficient Learning of Deep Networks from Decentralized Data
        H. Brendan McMahan, Eider Moore, Daniel Ramage, Seth Hampson, Blaise Agueray Arcas. AISTATS 2017.
        https://arxiv.org/abs/1602.05629
      This is also recommended model by "Adaptive Federated Optimization. ICML 2020" (https://arxiv.org/pdf/2003.00295.pdf)
      Args:
        vocab_size: the size of the vocabulary, used as a dimension in the input embedding.
        sequence_length: the length of input sequences.
      Returns:
        An uncompiled `torch.nn.Module`.
      """

    def __init__(self, embedding_dim=8, vocab_size=80, hidden_size=256):
        super(Model, self).__init__()
        self.embeddings = nn.Embedding(num_embeddings=vocab_size, embedding_dim=embedding_dim, padding_idx=0)
        self.lstm = nn.LSTM(input_size=embedding_dim, hidden_size=hidden_size, num_layers=2, batch_first=True)
        self.fc = nn.Linear(hidden_size, vocab_size)

    def forward(self, input_seq):
        embeds = self.embeddings(input_seq)
        # Note that the order of mini-batch is random so there is no hidden relationship among batches.
        # So we do not input the previous batch's hidden state,
        # leaving the first hidden state zero `self.lstm(embeds, None)`
        self.lstm.flatten_parameters()
        lstm_out, _ = self.lstm(embeds)
        # use the final hidden state as the next character prediction
        # final_hidden_state = lstm_out[:, -1]
        lstm_out = lstm_out[:,-1,:]
        output = self.fc(lstm_out)
        # For fed_shakespeare
        # output = self.fc(lstm_out[:,:])
        # output = torch.transpose(output, 1, 2)
        return output

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)


import os
import torch
import torch.utils.data
from typing import *
from torch import Tensor
from flgo.benchmark.base import FromDatasetPipe, FromDatasetGenerator, BasicTaskCalculator
try:
    import ujson as json
except:
    import json
from .config import train_data
try:
    from .config import test_data
except:
    test_data = None
try:
    from .config import val_data
except:
    val_data = None

class TaskGenerator(FromDatasetGenerator):
    def __init__(self, seq_len=5, stride=1):
        super(TaskGenerator, self).__init__(benchmark=os.path.split(os.path.dirname(__file__))[-1],
                                            train_data=train_data, val_data=val_data, test_data=test_data)
        self.seq_len = seq_len
        self.stride = stride

    def prepare_data_for_partition(self):
        return list(self.train_data)

def preprocess(data:List[Tensor], seq_len:int=5, stride:int=1):
    if data is None: return None
    all_data = torch.cat(data)
    seqs = [all_data[i:i+seq_len+1] for i in range(0, len(all_data)-seq_len-1, stride)]
    return torch.stack(seqs)

def collate_fn(batch):
    batch = torch.stack(batch)
    seq_len = batch.shape[-1]-1
    return batch[:,:seq_len], batch[:,1:]

class TaskPipe(FromDatasetPipe):
    class TaskDataset(torch.utils.data.Dataset):
        def __init__(self, data):
            if data is None: return None
            self.data = data

        def __len__(self):
            return len(self.data)

        def __getitem__(self, item):
            return self.data[item]

    def __init__(self, task_path):
        super(TaskPipe, self).__init__(task_path, train_data=train_data, val_data=val_data, test_data=test_data)

    def save_task(self, generator):
        client_names = self.gen_client_names(len(generator.local_datas))
        feddata = {'client_names': client_names, 'seq_len': generator.seq_len, 'stride':generator.stride}
        for cid in range(len(client_names)): feddata[client_names[cid]] = {'data': generator.local_datas[cid],}
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)
        return

    def load_data(self, running_time_option) -> dict:
        seq_len = self.feddata['seq_len']
        stride = self.feddata['stride']
        server_test_data = self.TaskDataset(preprocess(list(self.test_data) if self.test_data is not None else None, seq_len=seq_len, stride=stride))
        server_val_data = self.TaskDataset(preprocess(list(self.val_data) if self.val_data is not None else None, seq_len=seq_len, stride=stride))
        if server_test_data is not None and server_val_data is None:
            server_test_data, server_val_data = self.split_dataset(server_test_data, running_time_option['test_holdout'])
        # rearrange data for server
        task_data = {'server': {'test': server_test_data, 'val': server_val_data}}
        self.train_data = list(self.train_data)
        for cid, cname in enumerate(self.feddata['client_names']):
            cdata = [self.train_data[did] for did in self.feddata[cname]['data']]
            cdata = self.TaskDataset(preprocess(cdata, seq_len=seq_len, stride=stride))
            if running_time_option['train_holdout'] > 0:
                cdata_train, cdata_val = self.split_dataset(cdata, running_time_option['train_holdout'])
                if running_time_option['local_test']:
                    cdata_val, cdata_test = self.split_dataset(cdata_val, running_time_option['local_test_ratio'])
                else:
                    cdata_test = None
            else:
                cdata_train = cdata
                cdata_val, cdata_test = None, None
            task_data[cname] = {'train': cdata_train, 'val': cdata_val, 'test': cdata_test}
        return task_data

    def split_dataset(self, dataset, p=0.0):
        if dataset is None: return None, None
        if p==0.0: return dataset, None
        elif p==1.0: return None,dataset
        k = int(len(dataset)*p)
        return dataset[:k], dataset[k:]

class TaskCalculator(BasicTaskCalculator):
    def __init__(self, device, optimizer_name='sgd'):
        super(TaskCalculator, self).__init__(device, optimizer_name)
        self.DataLoader = torch.utils.data.DataLoader
        self.collate_fn = collate_fn
        self.criterion = torch.nn.CrossEntropyLoss()

    def compute_loss(self, model, data):
        """
        Args:
            model: the model to train
            data: the training dataset
        Returns: dict of train-one-step's result, which should at least contains the key 'loss'
        """
        sources, targets = self.to_device(data)
        outputs = model(sources)
        loss = self.criterion(outputs.view(-1, outputs.shape[-1]), targets.reshape(-1))
        return {'loss': loss}

    @torch.no_grad()
    def test(self, model, dataset, batch_size=64, num_workers=0, pin_memory=False):
        """
        Metric = [mean_accuracy, mean_loss]

        Args:
            model:
            dataset:
            batch_size:
        Returns: [mean_accuracy, mean_loss]
        """
        model.eval()
        if batch_size==-1:batch_size=len(dataset)
        data_loader = self.get_dataloader(dataset, batch_size=batch_size, num_workers=num_workers, pin_memory=pin_memory)
        total_loss = 0.0
        for batch_id, batch_data in enumerate(data_loader):
            batch_data = self.to_device(batch_data)
            outputs = model(batch_data[0])
            batch_mean_loss = self.criterion(outputs.view(-1, outputs.shape[-1]), batch_data[1].reshape(-1)).item()
            total_loss += batch_mean_loss*len(batch_data[0])
        total_loss = total_loss/len(dataset)
        return {'loss':total_loss}

    def to_device(self, data):
        return data[0].to(self.device), data[1].to(self.device)

    def get_dataloader(self, dataset, batch_size=64, shuffle=False, num_workers=0, pin_memory=False, drop_last=False):
        if self.DataLoader == None:
            raise NotImplementedError("DataLoader Not Found.")
        return self.DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, pin_memory=pin_memory, drop_last=drop_last, collate_fn=self.collate_fn)



import importlib
import shutil
from abc import ABCMeta, abstractmethod
import os
from typing import Callable
try:
    import ujson as json
except:
    import json
import torch
import random
from torch.utils.data import Dataset

class AbstractTaskGenerator(metaclass=ABCMeta):
    r"""
    Abstract Task Generator
    """
    @abstractmethod
    def load_data(self, *args, **kwarg):
        """Load the original data into memory that can be partitioned"""
        pass

    @abstractmethod
    def partition(self, *args, **kwarg):
        """Partition the loaded data into subsets of data owned by clients
        and the test data owned by the server
        """
        pass

    @abstractmethod
    def generate(self, *args, **kwarg):
        """Load and partition the data, and then generate the necessary
        information about the federated task (e.g. path, partition way, ...)"""
        pass

class AbstractTaskPipe(metaclass=ABCMeta):
    r"""
    Abstract Task Pipe
    """
    @abstractmethod
    def save_task(self, *args, **kwargs):
        """Save a federated task created by TaskGenerator as a static file on the disk"""
        pass

    @abstractmethod
    def load_task(self, *args, **kwargs):
        """Load a federated task from disk"""
        pass

class AbstractTaskCalculator(metaclass=ABCMeta):
    r"""
    Abstract Task Calculator
    """
    @abstractmethod
    def to_device(self, *args, **kwargs):
        """Put the data into the gpu device"""
        pass

    @abstractmethod
    def get_dataloader(self, *args, **kwargs):
        """Return a data loader that splits the input data into batches"""
        pass

    @abstractmethod
    def test(self, model, data, *args, **kwargs):
        """Evaluate the model on the data"""
        pass

    @abstractmethod
    def compute_loss(self, model, data, *args, **kwargs):
        """Compute the loss of the model on the data to complete the forward process"""
        pass

    @abstractmethod
    def get_optimizer(self, model, *args, **kwargs):
        """Return the optimizer on the parameters of the model"""
        pass

class BasicTaskGenerator(AbstractTaskGenerator):
    r"""
        Load the original dataset and partition the
        original dataset into local_movielens_recommendation data
    """
    def __init__(self, benchmark:str, rawdata_path:str):
        """
        Args:
            benchmark (str): the name of the federated task
            rawdata_path (str): the dictionary of the original dataset
        """
        # basic attribution
        self.benchmark = benchmark
        self.rawdata_path = rawdata_path
        # optional attribution
        self.partitioner = None
        self.train_data = None
        self.test_data = None
        self.val_data = None
        self.task_name = None
        self.para = {}
        self.additional_option = {}
        self.train_additional_option = {}
        self.test_additional_option = {}

    def generate(self, *args, **kwarg):
        """The whole process to generate federated task. """
        # load data
        self.load_data()
        # partition
        self.partition()
        # generate task name
        self.task_name = self.get_task_name()
        return

    def load_data(self, *args, **kwargs):
        """Download and load dataset into memory."""
        return

    def partition(self, *args, **kwargs):
        """Partition the data into different local_movielens_recommendation datasets"""
        return

    def register_partitioner(self, partitioner=None):
        """Register the partitioner as self's data partitioner"""
        self.partitioner = partitioner

    def init_para(self, para_list=None):
        pnames = list(self.para.keys())
        if para_list is not None:
            for i, pv in enumerate(para_list):
                pname = pnames[i]
                try:
                    self.para[pname] = type(self.para[pname])(pv)
                except:
                    self.para[pname] = pv
        for pname, pv in self.para.items():
            self.__setattr__(pname, pv)
        return

    def get_task_name(self):
        r"""
        Create the default name of the task
        """
        if not hasattr(self.partitioner, 'num_parties') and hasattr(self.partitioner, 'num_clients'):
            self.partitioner.num_parties = self.partitioner.num_clients
        else: self.partitioner.num_parties = 'unknown'
        return '_'.join(['B-' + self.benchmark, 'P-' + str(self.partitioner), 'N-' + str(self.partitioner.num_parties)])

class BasicTaskPipe(AbstractTaskPipe):
    r"""
    Store the partition information of TaskGenerator into the disk
    when generating federated tasks.

    Load the original dataset and the partition information to
    create the federated scenario when optimizing models
    """
    TaskDataset = None

    def __init__(self, task_path):
        r"""
        Args:
            task_path (str): the path of the federated task
        """
        self.task_path = task_path
        if os.path.exists(os.path.join(self.task_path, 'data.json')):
            with open(os.path.join(self.task_path, 'data.json'), 'r') as inf:
                self.feddata = json.load(inf)

    def save_task(self, generator):
        """Construct `feddata` and store it into the disk for recovering
        the partitioned datasets again from it"""
        raise NotImplementedError

    def load_data(self, running_time_option) -> dict:
        """Load the data and process it to the format that can be distributed
        to different objects"""
        raise NotImplementedError

    def generate_objects(self, running_time_option, algorithm, scene='horizontal') -> list:
        r"""
        Generate the virtual objects (i.e. coordinators and participants)
        in the FL system

        Args:
            running_time_option (dict): the option (i.e. configuration)
            algorithm (module|class): algorithm
            scene (str): horizontal or vertical
        """
        if scene=='horizontal':
            # init clients
            Client = algorithm.Client
            clients = [Client(running_time_option) for _ in range(len(self.feddata['client_names']))]
            for cid, c in enumerate(clients):
                c.id = cid
                c.name = self.feddata['client_names'][cid]
            # init server
            server = algorithm.Server(running_time_option)
            server.name = 'server'
            server.id = -1
            # bind clients and server
            server.register_clients(clients)
            for c in clients: c.register_server(server)
            # return objects as list
            objects = [server]
            objects.extend(clients)
        elif scene=='vertical':
            PassiveParty = algorithm.PassiveParty
            ActiveParty = algorithm.ActiveParty
            objects = []
            for pid, pname in enumerate(self.feddata['party_names']):
                is_active = self.feddata[pname]['data']['with_label']
                obj = ActiveParty(running_time_option) if is_active else PassiveParty(running_time_option)
                obj.id = pid
                obj.name = pname
                objects.append(obj)
            for party in objects:
                party.register_objects(objects)
        elif scene=='hierarchical':
            server = algorithm.Server(running_time_option)
            server.id = -1
            server.name = 'server'
            edge_servers = [algorithm.EdgeServer(running_time_option) for _ in range(self.feddata['num_edge_servers'])]
            for sid in range(len(edge_servers)):
                edge_servers[sid].id = sid
                edge_servers[sid].name = 'edge_server'+str(sid)
                edge_servers[sid].clients = []
            server.register_clients(edge_servers)
            clients = [algorithm.Client(running_time_option) for _ in range(len(self.feddata['client_names']))]
            edge_server_clients = [[] for _ in edge_servers]
            for cid, c in enumerate(clients):
                c.id = cid+len(edge_servers)
                c.name = self.feddata['client_names'][cid]
                edge_server_clients[self.feddata['client_group'][c.name]].append(c)
            for edge_server, client_set in zip(edge_servers,edge_server_clients):
                edge_server.register_clients(client_set)
            objects = [server]
            objects.extend(edge_servers)
            objects.extend(clients)
        elif scene=='decentralized':
            # init clients
            Client = algorithm.Client
            clients = [Client(running_time_option) for _ in range(len(self.feddata['client_names']))]
            for cid, c in enumerate(clients):
                c.id = cid
                c.name = self.feddata['client_names'][cid]
            # init topology of clients
            topology = self.feddata['topology']
            for c in clients:
                c.topology = topology
            adjacent = self.feddata['adjacent']
            for cid,c in enumerate(clients):
                c.clients = [clients[k] for k,nid in enumerate(adjacent[cid]) if nid>0]
            # init protocol
            protocol = algorithm.Protocol(running_time_option)
            protocol.name = 'protocol'
            # bind clients and server
            protocol.clients = clients
            # return objects as list
            objects = [protocol]
            objects.extend(clients)
        elif scene=='parallel_horizontal':
            # init clients
            Client = algorithm.Client
            clients = [Client(running_time_option) for _ in range(len(self.feddata['client_names']))]
            for cid, c in enumerate(clients):
                c.id = cid
                c.name = self.feddata['client_names'][cid]
            # init server
            server = algorithm.Server(running_time_option)
            server.name = 'server'
            server.id = -1
            # return objects as list
            objects = [server]
            objects.extend(clients)
        elif scene=='real_hserver':
            server = algorithm.Server(running_time_option)
            server.name = 'server'
            objects = [server]
        elif scene=='real_hclient':
            client = algorithm.Client(running_time_option)
            objects = [client]
        return objects

    def save_info(self, generator):
        r"""
        Save the basic information of the generated task into the disk
        """
        info = {'benchmark': '.'.join(generator.__module__.split('.')[:-1])}
        info['scene'] = generator.scene if hasattr(generator, 'scene') else 'unknown'
        info['num_clients'] = generator.num_clients if hasattr(generator, 'num_clients') else (generator.num_parties if hasattr(self, 'num_parties') else 'unknown')
        try:
            info['bmk_path'] = os.path.dirname(generator.__module__.__file__)
        except:
            pass
        if hasattr(generator, 'partitioner'):
            info['partitioner'] = str(generator.partitioner)
        with open(os.path.join(self.task_path, 'info'), 'w') as outf:
            json.dump(info, outf)

    def load_task(self, running_time_option, *args, **kwargs):
        r"""
        Load the generated task into disk and create objects in the federated
        scenario.
        """
        task_data = self.load_data(running_time_option)
        objects = self.generate_objects(running_time_option)
        self.distribute(task_data, objects)
        return objects

    def distribute(self, task_data: dict, objects: list):
        r"""
        Distribute the loaded local_movielens_recommendation datasets to different objects in
        the federated scenario
        """
        for ob in objects:
            if ob.name in task_data.keys():
                ob_data = task_data[ob.name]
                for data_name, data in ob_data.items():
                    ob.set_data(data, data_name)

    def split_dataset(self, dataset, p=0.0):
        r"""
        Split the dataset into two parts.

        Args:
            dataset (torch.utils.data.Dataset): the dataset to be splitted
            p (float): the ratio of the splitting

        Returns:
            The two split parts
        """
        if dataset is None: return None, None
        if p == 0 or len(dataset)<2: return dataset, None
        s1 = max(int(len(dataset) * p), 1)
        s2 = len(dataset) - s1
        if s1==0:
            return dataset, None
        elif s2==0:
            return None, dataset
        else:
            return torch.utils.data.random_split(dataset, [s2, s1])

    def task_exists(self):
        r"""
        Check whether the task already exists.

        Returns:
            True if the task already exists
        """
        return os.path.exists(self.task_path)

    def remove_task(self):
        r"""Remove this task"""
        if self.task_exists():
            shutil.rmtree(self.task_path)
        return

    def create_task_architecture(self):
        """Create the directories of the task."""
        if not self.task_exists():
            os.mkdir(self.task_path)
            os.mkdir(os.path.join(self.task_path, 'record'))
            os.mkdir(os.path.join(self.task_path, 'log'))
        else:
            raise FileExistsError("federated task {} already exists!".format(self.task_path))

    def gen_client_names(self, num_clients):
        r"""
        Generate the names of clients

        Returns:
            a list of strings
        """
        return [('Client{:0>' + str(len(str(num_clients))) + 'd}').format(i) for i in range(num_clients)]

class BasicTaskCalculator(AbstractTaskCalculator):
    r"""
    Support task-specific computation when optimizing models, such
    as putting data into device, computing loss, evaluating models,
    and creating the data loader
    """

    def __init__(self, device, optimizer_name='sgd'):
        r"""
        Args:
            device (torch.device): device
            optimizer_name (str): the name of the optimizer
        """
        self.device = device
        self.optimizer_name = optimizer_name
        self.criterion = None
        self.DataLoader = None
        self.collate_fn = None

    def to_device(self, data, *args, **kwargs):
        return NotImplementedError

    def get_dataloader(self, dataset, batch_size=64, *args, **kwargs):
        return NotImplementedError

    def test(self, model, data, *args, **kwargs):
        return NotImplementedError

    def compute_loss(self, model, data, *args, **kwargs):
        return NotImplementedError

    def get_optimizer(self, model=None, lr=0.1, weight_decay=0, momentum=0):
        r"""
        Create optimizer of the model parameters

        Args:
            model (torch.nn.Module): model
            lr (float): learning rate
            weight_decay (float): the weight_decay coefficient
            momentum (float): the momentum coefficient

        Returns:
            the optimizer
        """
        OPTIM = getattr(importlib.import_module('torch.optim'), self.optimizer_name)
        filter_fn = filter(lambda p: p.requires_grad, model.parameters())
        if self.optimizer_name.lower() == 'sgd':
            return OPTIM(filter_fn, lr=lr, momentum=momentum, weight_decay=weight_decay)
        elif self.optimizer_name.lower() in ['adam', 'rmsprop', 'adagrad']:
            return OPTIM(filter_fn, lr=lr, weight_decay=weight_decay)
        else:
            raise RuntimeError("Invalid Optimizer.")

    def set_criterion(self, criterion:Callable)->None:
        self.criterion = criterion

    def set_collate_fn(self, collate_fn:Callable)->None:
        self.collate_fn = collate_fn

class XYHorizontalTaskPipe(BasicTaskPipe):
    """
    This pipe is for supervised learning where each sample contains a feature $x_i$ and a label $y_i$
     that can be indexed by $i$.
    To use this pipe, it's necessary to set the attribute `test_data` of the generator to be a dict like:
        {'x': [...], 'y':[...]}
    and the attribute `local_datas` to be a list of the above dict that means the local_movielens_recommendation data owned by clients:
        [{'x':[...], 'y':[...]}, ..., ]
    """
    TaskDataset = torch.utils.data.TensorDataset

    def save_task(self, generator):
        client_names = self.gen_client_names(len(generator.local_datas))
        feddata = {'client_names': client_names, 'server': {'data': generator.test_data}}
        for cid in range(len(client_names)): feddata[client_names[cid]] = {'data': generator.local_datas[cid]}
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)

    def load_data(self, running_time_option) -> dict:
        test_data = self.feddata['server']['data']
        test_data = self.TaskDataset(torch.tensor(test_data['x']), torch.tensor(test_data['y']))
        local_datas = [self.TaskDataset(torch.tensor(self.feddata[cname]['data']['x']),
                                        torch.tensor(self.feddata[cname]['data']['y'])) for cname in
                       self.feddata['client_names']]
        server_data_test, server_data_val = self.split_dataset(test_data, running_time_option['test_holdout'])
        task_data = {'server': {'test': server_data_test, 'val': server_data_val}}
        for key in self.feddata['server'].keys():
            if key == 'data':
                continue
            task_data['server'][key] = self.feddata['server'][key]
        for cid, cname in enumerate(self.feddata['client_names']):
            cdata = local_datas[cid]
            cdata_train, cdata_val = self.split_dataset(cdata, running_time_option['train_holdout'])
            if running_time_option['local_test'] and cdata_val is not None:
                cdata_val, cdata_test = self.split_dataset(cdata_val, running_time_option['local_test_ratio'])
            else:
                cdata_test = None
            task_data[cname] = {'train': cdata_train, 'val': cdata_val, 'test': cdata_test}
            for key in self.feddata[cname]:
                if key == 'data':
                    continue
                task_data[cname][key] = self.feddata[cname][key]
        return task_data

class XHorizontalTaskPipe(BasicTaskPipe):
    TaskDataset = torch.utils.data.TensorDataset
    def save_task(self, generator):
        client_names = self.gen_client_names(len(generator.local_datas))
        feddata = {'client_names': client_names, 'server': {'data': generator.test_data}}
        for cid in range(len(client_names)): feddata[client_names[cid]] = {'data': generator.local_datas[cid]}
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)

    def load_data(self, running_time_option) -> dict:
        test_data = self.feddata['server']['data']
        test_data = self.TaskDataset(torch.tensor(test_data['x']))
        local_datas = [self.TaskDataset(torch.tensor(self.feddata[cname]['data']['x']), ) for cname in self.feddata['client_names']]
        server_data_test, server_data_val = self.split_dataset(test_data, running_time_option['test_holdout'])
        task_data = {'server': {'test': server_data_test, 'val': server_data_val}}
        for key in self.feddata['server'].keys():
            if key == 'data':
                continue
            task_data['server'][key] = self.feddata['server'][key]
        for cid, cname in enumerate(self.feddata['client_names']):
            cdata = local_datas[cid]
            cdata_train, cdata_val = self.split_dataset(cdata, running_time_option['train_holdout'])
            if running_time_option['local_test'] and cdata_val is not None:
                cdata_val, cdata_test = self.split_dataset(cdata_val, running_time_option['local_test_ratio'])
            else:
                cdata_test = None
            task_data[cname] = {'train': cdata_train, 'val': cdata_val, 'test':cdata_test}
            for key in self.feddata[cname]:
                if key == 'data':
                    continue
                task_data[cname][key] = self.feddata[cname][key]
        return task_data

class FromDatasetGenerator(BasicTaskGenerator):
    r"""
    This generator will do:
    1. Directly create train_data and test_data from input;
    2. Convert the train_data into the scheme that can be partitioned by Partitioner if necessary.
    """
    def __init__(self, benchmark, train_data, val_data=None, test_data=None):
        super(FromDatasetGenerator, self).__init__(benchmark=benchmark, rawdata_path='')
        self.train_data = train_data
        self.val_data = val_data
        self.test_data = test_data

    def generate(self, *args, **kwarg):
        self.partition()

    def partition(self, *args, **kwargs):
        self.train_data = self.prepare_data_for_partition()
        self.local_datas = self.partitioner(self.train_data)
        self.num_clients = len(self.local_datas)

    def prepare_data_for_partition(self):
        """Transform the attribution self.train_data into the format that can be received by partitioner"""
        return self.train_data

class FromDatasetPipe(BasicTaskPipe):
    TaskDataset = None
    def __init__(self, task_path, train_data, val_data = None, test_data=None):
        super(FromDatasetPipe, self).__init__(task_path)
        self.train_data = train_data
        self.val_data = val_data
        self.test_data = test_data

class DecentralizedFromDatasetPipe(FromDatasetPipe):
    TaskDataset = None
    def __init__(self, task_path, train_data, val_data = None, test_data=None):
        super(DecentralizedFromDatasetPipe, self).__init__(task_path, train_data, val_data, test_data)

    def save_topology(self, feddata:{}):
        client_names = feddata['client_names']
        feddata['topology'] = self.topology
        num_clients = len(client_names)
        if self.adjacent is None:
            if self.topology == 'mesh':
                feddata['adjacent'] = [[1.0 for __ in range(num_clients)] for _ in range(num_clients)]
                for i in range(num_clients):
                    feddata['adjacent'][i][i] = 0.0
            elif self.topology == 'line':
                feddata['adjacent'] = [[0.0 for __ in range(num_clients)] for _ in range(num_clients)]
                for i in range(1, num_clients):
                    feddata['adjacent'][i][((i - 1) % num_clients)] = 1.0
            elif self.topology == 'ring':
                feddata['adjacent'] = [[0.0 for __ in range(num_clients)] for _ in range(num_clients)]
                for i in range(num_clients):
                    feddata['adjacent'][i][((i - 1) % num_clients)] = 1.0
            elif self.topology == 'random':
                feddata['adjacent'] = [[0.0 for __ in range(num_clients)] for _ in range(num_clients)]
                p = 0.5
                for i in range(num_clients):
                    for k in range(num_clients):
                        if i == k: continue
                        if random.random() < p:
                            feddata['adjacent'][i][k] = 1.0
        return feddata

class HierFromDatasetGenerator(FromDatasetGenerator):
    def partition(self, *args, **kwargs):
        self.train_data = self.prepare_data_for_partition()
        self.local_datas = self.partitioner(self.train_data)
        self.num_edge_servers = len(self.local_datas)
        self.num_clients = sum([len(d) for d in self.local_datas])

class HierFromDatasetPipe(FromDatasetPipe):
    def create_feddata(self, generator):
        num_clients = sum([len(d) for d in generator.local_datas])
        client_names = self.gen_client_names(num_clients)
        feddata = {'client_names': client_names}
        feddata['client_group'] = {}
        k = 0
        client_data = {cname:{} for cname in client_names}
        for gid in range(len(generator.local_datas)):
            group_data = generator.local_datas[gid]
            for i in range(len(group_data)):
                feddata['client_group'][client_names[k+i]] = gid
                client_data[client_names[k+i]]['data'] = generator.local_datas[gid][i]
            k+=len(group_data)
        feddata['num_edge_servers'] = len(generator.local_datas)
        feddata.update(client_data)
        return feddata

    def save_task(self, generator):
        feddata = self.create_feddata(generator)
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)
        return


"""
This is a non-official implementation of 'Federated Learning Based on Dynamic Regularization'
(http://arxiv.org/abs/2111.04263). The official implementation is at 'https://github.com/alpemreacar/FedDyn'
"""
from .fedbase import BasicServer, BasicClient
import copy
from flgo.utils import fmodule
import torch

class Server(BasicServer):
    def initialize(self, *args, **kwargs):
        self.init_algo_para({'alpha': 0.1})
        self.h = self.model.zeros_like()

    def aggregate(self, models):
        if len(models) == 0: return self.model
        nan_exists = [m.has_nan() for m in models]
        if any(nan_exists):
            if all(nan_exists): raise ValueError("All the received local models have parameters of nan value.")
            self.gv.logger.info('Warning("There exists nan-value in local models, which will be automatically removed from the aggregatino list.")')
            new_models = []
            received_clients = []
            for ni, mi, cid in zip(nan_exists, models, self.received_clients):
                if ni: continue
                new_models.append(mi)
                received_clients.append(cid)
            self.received_clients = received_clients
            models = new_models
        self.h = self.h - self.alpha * 1.0 / self.num_clients * (fmodule._model_sum(models) - len(models)*self.model)
        new_model = fmodule._model_average(models) - 1.0 / self.alpha * self.h
        return new_model

class Client(BasicClient):
    def initialize(self, *args, **kwargs):
        self.gradL = None

    @ fmodule.with_multi_gpus
    def train(self, model):
        if self.gradL == None:self.gradL = model.zeros_like()
        self.gradL.to(model.get_device())
        # global parameters
        src_model = copy.deepcopy(model)
        src_model.freeze_grad()
        model.train()
        optimizer = self.calculator.get_optimizer(model, lr=self.learning_rate, weight_decay=self.weight_decay, momentum=self.momentum)
        for iter in range(self.num_steps):
            batch_data = self.get_batch_data()
            model.zero_grad()
            l1 = self.calculator.compute_loss(model, batch_data)['loss']
            l2 = 0
            l3 = 0
            for pgl, pm, ps in zip(self.gradL.parameters(), model.parameters(), src_model.parameters()):
                l2 += torch.dot(pgl.view(-1), pm.view(-1))
                l3 += torch.sum(torch.pow(pm - ps, 2))
            loss = l1 - l2 + 0.5 * self.alpha * l3
            loss.backward()
            if self.clip_grad>0:torch.nn.utils.clip_grad_norm_(parameters=model.parameters(), max_norm=self.clip_grad)
            optimizer.step()
        # update grad_L
        self.gradL = self.gradL - self.alpha * (model-src_model)
        self.gradL.to(torch.device('cpu'))
        return


import flgo.benchmark.toolkits.visualization
from .model import fcn_resnet50, unet
import flgo.benchmark.toolkits.partition

default_partitioner = flgo.benchmark.toolkits.partition.IIDPartitioner
default_partition_para = {'num_clients':100}
default_model = fcn_resnet50



from flgo.experiment.logger import BasicLogger
import numpy as np

class FullLogger(BasicLogger):
    def log_once(self, *args, **kwargs):
        test_metric = self.server.test()
        for met_name, met_val in test_metric.items():
            self.output['test_' + met_name].append(met_val)
        val_metric = self.server.test(flag='val')
        for met_name, met_val in val_metric.items():
            self.output['val_' + met_name].append(met_val)
        # calculate weighted averaging of metrics on training datasets across participants
        local_data_vols = [c.datavol for c in self.clients]
        total_data_vol = sum(local_data_vols)
        train_metrics = self.server.global_test(flag='train')
        for met_name, met_val in train_metrics.items():
            self.output['train_' + met_name + '_dist'].append(met_val)
            self.output['train_' + met_name].append(1.0 * sum([client_vol * client_met for client_vol, client_met in zip(local_data_vols, met_val)]) / total_data_vol)
        # calculate weighted averaging and other statistics of metrics on validation datasets across clients
        local_val_metrics = self.server.global_test(flag='val')
        for met_name, met_val in local_val_metrics.items():
            self.output['local_val_'+met_name+'_dist'].append(met_val)
            self.output['local_val_' + met_name].append(1.0 * sum([client_vol * client_met for client_vol, client_met in zip(local_data_vols, met_val)]) / total_data_vol)
            self.output['mean_local_val_' + met_name].append(np.mean(met_val))
            self.output['std_local_val_' + met_name].append(np.std(met_val))
        local_test_metrics = self.server.global_test(flag='test')
        for met_name, met_val in local_test_metrics.items():
            self.output['local_test_'+met_name+'_dist'].append(met_val)
            self.output['local_test_' + met_name].append(1.0 * sum([client_vol * client_met for client_vol, client_met in zip(local_data_vols, met_val)]) / total_data_vol)
            self.output['mean_local_test_' + met_name].append(np.mean(met_val))
            self.output['std_local_test_' + met_name].append(np.std(met_val))
        # output to stdout
        self.show_current_output()


import torch.nn
import flgo.benchmark
import os
import torch_geometric
import torch_geometric.transforms as T
from torch_geometric.nn import GCNConv
from torch_geometric.utils import negative_sampling
import torch

class GCN(torch.nn.Module):
    def __init__(self):
        super(GCN, self).__init__()
        self.conv1 = GCNConv(500, 128)
        self.conv2 = GCNConv(128, 64)

    def forward(self, tdata):
        z = self.encode(tdata.x, tdata.edge_index)
        if self.training:
            neg_edge_index = negative_sampling(
                edge_index=torch.cat([tdata.edge_index, tdata.edge_label_index],dim=-1,), num_nodes=tdata.num_nodes,
                num_neg_samples=tdata.edge_label_index.size(1),
                force_undirected=tdata.is_undirected()
            )
            outputs = self.decode(z, torch.cat([tdata.edge_label_index, neg_edge_index],dim=-1,)).view(-1)
            return outputs, neg_edge_index
        else:
            outputs = self.decode(z, tdata.edge_label_index).view(-1)
            return outputs

    def encode(self, x, edge_index):
        x = self.conv1(x, edge_index.long())
        x = x.relu()
        return self.conv2(x, edge_index.long())

    def decode(self, z, edge_label_index):
        return (z[edge_label_index.long()[0]] * z[edge_label_index.long()[1]]).sum(dim=-1)

    def decode_all(self, z):
        prob_adj = z @ z.t()
        return (prob_adj > 0).nonzero(as_tuple=False).t()

trans = T.NormalizeFeatures()
path = os.path.join(flgo.benchmark.data_root, 'PUBMED')
dataset = torch_geometric.datasets.Planetoid(path, name='Pubmed', transform=trans)
train_data = dataset[0]

def get_model():
    return GCN()


from flgo.algorithm.fedbase import BasicParty
import collections
from abc import ABCMeta, abstractmethod
from flgo.utils import fmodule
import copy
import torch

class AbstractProtocol(metaclass=ABCMeta):
    r"""
    Abstract Protocol
    """
    @abstractmethod
    def get_clients_for_iteration(self, *args, **kwarg):
        """Return clients that should perform iteration at the current moment"""
        pass

class BasicProtocol(AbstractProtocol, BasicParty):
    def __init__(self, option):
        super().__init__()
        self.test_data = None
        self.val_data = None
        self.model = None
        self.clients = []
        # basic configuration
        self.task = option['task']
        self.eval_interval = option['eval_interval']
        self.num_parallels = option['num_parallels']
        # server calculator
        self.device = self.gv.apply_for_device() if not option['server_with_cpu'] else torch.device('cpu')
        self.calculator = self.gv.TaskCalculator(self.device, optimizer_name=option['optimizer'])
        # hyper-parameters during training process
        self.num_rounds = option['num_rounds']
        # algorithm-dependent parameters
        self.algo_para = {}
        self.current_round = 1
        # all options
        self.option = option
        self.id = -1

    def get_clients_for_iteration(self, *args, **kwarg):
        return self.clients

    def run(self):
        self.gv.logger.time_start('Total Time Cost')
        if self.eval_interval > 0:
            self.gv.logger.info("--------------Initial Evaluation--------------")
            self.gv.logger.time_start('Eval Time Cost')
            self.gv.logger.log_once()
            self.gv.logger.time_end('Eval Time Cost')
        for c in self.clients:
            c.init_state()
        while self.current_round <= self.num_rounds:
            updated = self.iterate()
            # using logger to evaluate the model if the model is updated
            if updated is True or updated is None:
                self.gv.logger.info("--------------Round {}--------------".format(self.current_round))
                # check log interval
                if self.gv.logger.check_if_log(self.current_round, self.eval_interval):
                    self.gv.logger.time_start('Eval Time Cost')
                    self.gv.logger.log_once()
                    self.gv.logger.time_end('Eval Time Cost')
                # check if early stopping
                if self.gv.logger.early_stop(): break
                self.current_round += 1
        self.gv.logger.info("=================End==================")
        self.gv.logger.time_end('Total Time Cost')
        # save results as .json file
        self.gv.logger.save_output_as_json()
        return

    def iterate(self):
        clients_for_iteration = self.get_clients_for_iteration()
        # compute next state for each client in Sc
        for c in clients_for_iteration:
            c.nstate = c.update_state()
        # flush the states of all the clients
        for c in clients_for_iteration:
            c.state = copy.deepcopy(c.nstate)
        return

class LineProtocol(BasicProtocol):
    def __init__(self, option):
        super(LineProtocol, self).__init__(option)
        self.crt_client = 0
        self.num_rounds = min(len(self.clients), self.num_rounds)

    def get_clients_for_iteration(self, *args, **kwarg):
        ret = [self.clients[self.crt_client]]
        self.crt_client += 1
        return ret

class RingProtocol(BasicProtocol):
    def __init__(self, option):
        super(RingProtocol, self).__init__(option)
        self.crt_client = 0

    def get_clients_for_iteration(self, *args, **kwarg):
        ret = [self.clients[self.crt_client]]
        self.crt_client = (self.crt_client+1)%len(self.clients)
        return ret

class BasicClient(BasicParty):
    def __init__(self, option):
        super().__init__(option)
        self.clients = None
        self.topology = 'mesh'
        self.model = None
        self.state = {}
        self.pstate = {}
        self.nstate = {}
        self.id = None
        self.data_loader = None
        self.test_data = None
        self.val_data = None
        self.train_data = None
        self.device = self.gv.apply_for_device()
        self.calculator = self.gv.TaskCalculator(self.device, option['optimizer'])
        self._train_loader = None
        # hyper-parameters for training
        self.optimizer_name = option['optimizer']
        self.learning_rate = option['learning_rate']
        self.momentum = option['momentum']
        self.weight_decay = option['weight_decay']
        self.batch_size = option['batch_size']
        self.num_steps = option['num_steps']
        self.num_epochs = option['num_epochs']
        self.clip_grad = option['clip_grad']
        self.model = None
        self.test_batch_size = option['test_batch_size']
        self.loader_num_workers = option['num_workers']
        self.current_steps = 0
        # system setting
        self._effective_num_steps = self.num_steps
        self._latency = 0
        # actions of different message type
        self.option = option
        self.actions = {0: self.send_state}

    def init_state(self):
        # self.state will be shared with other clients and any direct change on self.state should not impact self.model
        self.state = copy.deepcopy({'model': self.model})

    def update_state(self):
        # sample clients
        selected_clients = self.sample()
        # request prior states from selected neighbors
        received_models = self.communicate(selected_clients)['model']
        # aggregate neighbors' states
        self.model = self.aggregate(received_models)
        # local train
        self.train(self.model)
        return {'model': self.model}

    def sample(self):
        if self.topology=='mesh':
            return list(range(len(self.clients)))
        elif self.topology=='line' or self.topology=='ring':
            return [0]
        else:
            raise NotImplementedError('Not Supported topology')

    def communicate(self, selected_clients, mtype=0, asynchronous=False):
        packages_received_from_clients = []
        received_package_buffer = {}
        communicate_clients = selected_clients
        # prepare packages for clients
        for client_id in communicate_clients:
            received_package_buffer[client_id] = None
        # communicate with selected clients
        for client_id in communicate_clients:
            server_pkg = self.pack(client_id, mtype)
            server_pkg['__mtype__'] = mtype
            response_from_client_id = self.communicate_with(self.clients[client_id].id, package=server_pkg)
            packages_received_from_clients.append(response_from_client_id)
        for i, client_id in enumerate(communicate_clients): received_package_buffer[client_id] = packages_received_from_clients[i]
        packages_received_from_clients = [received_package_buffer[cid] for cid in selected_clients if
                                          received_package_buffer[cid]]
        self.received_clients = selected_clients
        return self.unpack(packages_received_from_clients)

    def communicate_with(self, target_id, package={}):
        return self.gv.communicator.request(self.id, target_id, package)

    def aggregate(self, models:list):
        return fmodule._model_average(models+[self.model])

    def train(self, model):
        r"""
        Standard local_movielens_recommendation training procedure. Train the transmitted model with
        local_movielens_recommendation training dataset.

        Args:
            model (FModule): the global model
        """
        model.train()
        optimizer = self.calculator.get_optimizer(model, lr=self.learning_rate, weight_decay=self.weight_decay,
                                                  momentum=self.momentum)
        for iter in range(self.num_steps):
            # get a batch of data
            batch_data = self.get_batch_data()
            model.zero_grad()
            # calculate the loss of the model on batched dataset through task-specified calculator
            loss = self.calculator.compute_loss(model, batch_data)['loss']
            loss.backward()
            if self.clip_grad>0:torch.nn.utils.clip_grad_norm_(parameters=model.parameters(), max_norm=self.clip_grad)
            optimizer.step()
        return

    def get_batch_data(self):
        """
        Get the batch of training data
        Returns:
            a batch of data
        """
        if self._train_loader is None:
            self._train_loader = self.calculator.get_dataloader(self.train_data, batch_size=self.batch_size,
                                                                   num_workers=self.loader_num_workers,
                                                                   pin_memory=self.option['pin_memory'], drop_last = self.option.get('drop_last', False))
        try:
            batch_data = next(self.data_loader)
        except Exception as e:
            self.data_loader = iter(self._train_loader)
            batch_data = next(self.data_loader)
        # clear local_movielens_recommendation DataLoader when finishing local_movielens_recommendation training
        self.current_steps = (self.current_steps + 1) % self.num_steps
        if self.current_steps == 0:
            self.data_loader = None
            self._train_loader = None
        return batch_data

    def pack(self, *args, **kwargs):
        return {}

    def unpack(self, packages_received_from_clients):
        r"""
        Unpack the information from the received packages. Return models and losses as default.

        Args:
            packages_received_from_clients (list): a list of packages

        Returns:
            res (dict): collections.defaultdict that contains several lists of the clients' reply
        """
        if len(packages_received_from_clients) == 0: return collections.defaultdict(list)
        res = {pname: [] for pname in packages_received_from_clients[0]}
        for cpkg in packages_received_from_clients:
            for pname, pval in cpkg.items():
                res[pname].append(pval)
        return res

    def send_state(self, *args, **kwargs):
        return copy.deepcopy(self.state)

    def test(self, model=None, flag='val'):
        r"""
        Evaluate the model on the dataset owned by the client

        Args:
            model (flgo.utils.fmodule.FModule): the model need to be evaluated
            flag (str): choose the data to evaluate the model

        Returns:
            metric (dict): the evaluating results (e.g. metric = {'loss':1.02})
        """
        if model is None: model = self.model
        dataset = getattr(self, flag + '_data') if hasattr(self, flag + '_data') else None
        if dataset is None: return {}
        return self.calculator.test(model, dataset, min(self.test_batch_size, len(dataset)), self.option['num_workers'])

Client = BasicClient
Protocol = BasicProtocol


from torch import nn
from flgo.utils.fmodule import FModule

num_users = 20928
num_items = 295289

class Model(FModule):
    def __init__(self, n_factors=10):
        super().__init__()
        self.user_factors = nn.Embedding(num_users, n_factors)
        self.item_factors = nn.Embedding(num_items, n_factors)
        self.sigmoid = nn.Sigmoid()

    def forward(self, batch_data):
        user_id = batch_data['user_id'].long()
        item_id = batch_data['item_id'].long()
        u, it = self.user_factors(user_id), self.item_factors(item_id)
        x = self.sigmoid((u * it).sum(1))
        return x * 5

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        global num_users, num_items
        num_users = object.test_data.num_users
        num_items = object.test_data.num_items
        object.model = Model().to(object.device)


from .utils.fflow import init, gen_task, gen_task_by_, gen_benchmark_from_file, gen_direct_task_from_file, zip_task, pull_task_from_, gen_real_task, gen_decentralized_benchmark, gen_hierarchical_benchmark, gen_empty_task, convert_model,tune, run_in_parallel, module2fmodule, multi_init_and_run, set_data_root,download_resource, list_resource, option_helper
from .benchmark import data_root
communicator = None
_data_root = data_root
_name = None
__version__ = "v0.3.0"
class VirtualCommunicator:
    """
    Communicator that simulates the communication phase between any two objects
    """
    def __init__(self, objects):
        self.objects_map = {obj.id:obj for obj in objects}
        self.objects = objects

    def request(self, source, target, package):
        # send package to the target object with `package` and `mtype`, and then listen from it
        return self.objects_map[target].message_handler(package)



import os
import random
import torch
import torch_geometric.transforms as T
from flgo.benchmark.toolkits.graph.node_classification import GeneralCalculator, FromDatasetPipe, FromDatasetGenerator
from .config import train_data
try:
    import ujson as json
except:
    import json
try:
    from .config import test_data
except:
    test_data = None
try:
    from .config import val_data
except:
    val_data = None

class TaskGenerator(FromDatasetGenerator):
    def __init__(self, transductive=True):
        super(TaskGenerator, self).__init__(benchmark=os.path.split(os.path.dirname(__file__))[-1],
                                            train_data=train_data, val_data=val_data, test_data=test_data)
        self.transductive = transductive

class TaskPipe(FromDatasetPipe):
    def __init__(self, task_path):
        super(TaskPipe, self).__init__(task_path, train_data=train_data, val_data=val_data, test_data=test_data)

    def save_task(self, generator):
        client_names = self.gen_client_names(len(generator.local_datas))
        feddata = {'client_names': client_names, 'transductive': generator.transductive}
        for cid in range(len(client_names)):
            feddata[client_names[cid]] = {'data': generator.local_datas[cid], }
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)
        return

    def load_data(self, running_time_option) -> dict:
        transductive = self.feddata['transductive']
        if self.test_data is not None:
            if self.val_data is None:
                if running_time_option['test_holdout']>0:
                    all_nodes = list(range(self.test_data.num_nodes))
                    random.shuffle(all_nodes)
                    k = int(self.test_data.num_nodes * running_time_option['test_holdout'])
                    val_nodes = all_nodes[:k]
                    test_nodes = all_nodes[k:]
                    if transductive:
                        test_mask = torch.BoolTensor([0 for _ in range(len(all_nodes))])
                        val_mask = torch.BoolTensor([0 for _ in range(len(all_nodes))])
                        val_mask[val_nodes] = True
                        test_mask[test_nodes] = True
                        self.test_data.val_mask = val_mask
                        self.test_data.test_mask = test_mask
                        server_test_data = self.TaskDataset(self.test_data, 'test')
                        server_val_data = self.TaskDataset(self.test_data, 'val')
                    else:
                        val_mask = torch.BoolTensor([0 for _ in range(len(all_nodes))])
                        val_mask[val_nodes] = True
                        val_data = self.test_data.subgraph(val_mask)
                        val_data.val_mask = torch.BoolTensor([1 for _ in range(val_data.num_nodes)])
                        test_mask = torch.BoolTensor([0 for _ in range(len(all_nodes))])
                        test_mask[test_nodes] = True
                        self.test_data.tesk_mask = test_mask
                        server_val_data = self.TaskDataset(val_data, 'val')
                        server_test_data = self.TaskDataset(self.test_data, 'test')
            else:
                self.test_data.test_mask = torch.BoolTensor([1 for _ in range(self.test_data.num_nodes)])
                self.val_data.val_mask = torch.BoolTensor([1 for _ in range(self.val_data.num_nodes)])
                server_test_data = self.TaskDataset(self.test_data, 'test')
                server_val_data = self.TaskDataset(self.val_data, 'val')
        elif self.val_data is not None:
            self.val_data.val_mask = torch.BoolTensor([1 for _ in range(self.val_data.num_nodes)])
            server_test_data = None
            server_val_data = self.TaskDataset(self.val_data, 'val')
        else:
            server_test_data = server_val_data = None
        task_data = {'server': {'test':server_test_data, 'val':server_val_data}}
        # rearrange data for clients
        num_val = running_time_option['train_holdout']
        if running_time_option['local_test']:
            num_val = 0.5 * num_val
            num_test = num_val
        else:
            num_test = 0.0
        for cid, cname in enumerate(self.feddata['client_names']):
            cnodes = self.feddata[cname]['data']
            cmask = torch.BoolTensor([0 for _ in range(self.train_data.num_nodes)])
            cmask[cnodes] = True
            cdata = self.train_data.subgraph(cmask)
            cdata = T.RandomNodeSplit(num_val=num_val, num_test=num_test)(cdata)
            if transductive:
                task_data[cname] = {'train': self.TaskDataset(cdata, 'train'), 'val': self.TaskDataset(cdata, 'val') if torch.any(cdata.val_mask) else None, 'test':self.TaskDataset(cdata, 'test') if torch.any(cdata.test_mask) else None}
            else:
                c_local_nodes = list(range(cdata.num_nodes))
                random.shuffle(c_local_nodes)
                ctrain_data = cdata.subgraph(cdata.train_mask)
                cval_data = cdata.subgraph(cdata.train_mask+cdata.val_mask)
                ctest_data = cdata
                task_data[cname] = {'train': self.TaskDataset(ctrain_data, 'train'), 'val': self.TaskDataset(cval_data, 'val') if num_val>0 else None, 'test': self.TaskDataset(ctest_data,'test') if num_test>0 else None}
        return task_data


TaskCalculator = GeneralCalculator


from ..toolkits.rec.datasets import MovieLens
from ..toolkits.rec.rating_prediction import BuiltinClassGenerator, BuiltinClassPipe, GeneralCalculator
import os
import flgo.benchmark

path = os.path.join(flgo.benchmark.data_root, 'MOVIELENS')
builtin_class  = MovieLens

class TaskGenerator(BuiltinClassGenerator):
    def __init__(self, rawdata_path=path, split='100k', min_val=10, max_val=10e8):
        super(TaskGenerator, self).__init__(os.path.split(os.path.dirname(__file__))[-1], rawdata_path=rawdata_path, builtin_class=builtin_class)
        self.additional_option = {'split': split, 'download':True, 'min_val':min_val, 'max_val':max_val}

class TaskPipe(BuiltinClassPipe):
    def __init__(self, task_path):
        super(TaskPipe, self).__init__(task_path, builtin_class)

TaskCalculator = GeneralCalculator


import urllib
import zipfile
import torch
from torch.utils.data import TensorDataset, Dataset
from flgo.benchmark.toolkits import BasicTaskGenerator, BasicTaskCalculator
from flgo.benchmark.base import BasicTaskPipe
import collections
import re
import os
import os.path
try:
    import ujson as json
except:
    import json
import flgo.benchmark
import os.path
def download_from_url(url= None, filepath = '.'):
    """Download dataset from url to filepath."""
    if url:urllib.request.urlretrieve(url, filepath)
    return filepath

def extract_from_zip(src_path, target_path):
    """Unzip the .zip file (src_path) to target_path"""
    f = zipfile.ZipFile(src_path)
    f.extractall(target_path)
    targets = f.namelist()
    f.close()
    return [os.path.join(target_path, tar) for tar in targets]

class SHAKESPEARE(Dataset):
    def __init__(self, root, train=True):

        self.train = train
        self.root = root
        self.CHARACTER_RE = re.compile(r'^  ([a-zA-Z][a-zA-Z ]*)\. (.*)')
        self.CONT_RE = re.compile(r'^    (.*)')
        # The Comedy of Errors has errors in its indentation so we need to use
        # different regular expressions.
        self.COE_CHARACTER_RE = re.compile(r'^([a-zA-Z][a-zA-Z ]*)\. (.*)')
        self.COE_CONT_RE = re.compile(r'^(.*)')
        self.ALL_LETTERS = "\n !\"&'(),-.0123456789:;>?ABCDEFGHIJKLMNOPQRSTUVWXYZ[]abcdefghijklmnopqrstuvwxyz}"
        self.NUM_LETTERS = len(self.ALL_LETTERS)
        self.SEQ_LENGTH = 80

        file = 'train_data.json' if self.train else 'test_data.json'
        if not os.path.exists(os.path.join(self.root, 'raw_data', file)):
            self.download()
        with open(os.path.join(self.root, 'raw_data', file), 'r') as f:
            data = json.load(f)
            if self.train:
                self.id = data['id']
            self.x = data['x']
            self.y = data['y']

    def __getitem__(self, index):
        return torch.tensor(self.x[index], dtype=torch.long), self.y[index]

    def __len__(self):
        return len(self.x)

    def download(self):
        raw_path = os.path.join(self.root, 'raw_data')
        if not os.path.exists(os.path.join(raw_path, 'all_data.json')):
            if not os.path.exists(raw_path):
                os.makedirs(raw_path)
            print('Downloading...')
            src_path = download_from_url("https://www.gutenberg.org/files/100/old/1994-01-100.zip",
                                         os.path.join(raw_path, 'tmp'))
            tar_paths = extract_from_zip(src_path, raw_path)
            os.remove(src_path)

            with open(tar_paths[0], 'r') as input_file:
                shakespeare_full = input_file.read()
            plays, discarded_lines = self._split_into_plays(shakespeare_full)
            users_and_plays, all_examples, num_skipped = self._get_examples_by_character(plays)
            all_data = {}
            for user in all_examples.keys():
                all_data[user] = {
                    'play': users_and_plays[user],
                    'sound_bites': all_examples[user],
                }
            with open(os.path.join(self.root, 'raw_data', 'all_data.json'), 'w') as f:
                json.dump(all_data, f)
            os.remove(tar_paths[0])
        else:
            with open(os.path.join(self.root, 'raw_data', 'all_data.json'), 'r') as f:
                all_data = json.load(f)
        print('Processing...')
        self.process(all_data)

    def process(self, all_data):
        users = list(all_data.keys())
        num_user = len(users)
        s1 = int(num_user * 0.9)
        s2 = num_user - s1
        train_users, test_users = torch.utils.data.random_split(users, [s1, s2])
        train_sample_ids = []
        trainXs = []
        trainYs = []
        testXs = []
        testYs = []
        cid = 0
        for uid in train_users.indices:
            user = users[uid]
            examples = all_data[user]['sound_bites']
            X, Y = self.example_to_text(examples)
            X = self.X_text_to_vec(X)
            Y = self.Y_text_to_vec(Y)
            if len(X) < 10:
                testXs.extend(X)
                testYs.extend(Y)
                continue
            trainXs.extend(X)
            trainYs.extend(Y)
            train_sample_ids.extend([cid] * len(X))
            cid += 1
        train_data = {
            'x': trainXs,
            'y': trainYs,
            'id': train_sample_ids
        }
        with open(os.path.join(self.root, 'raw_data', 'train_data.json'), 'w') as f:
            json.dump(train_data, f)
        for uid in test_users.indices:
            user = list(all_data.keys())[uid]
            examples = all_data[user]['sound_bites']
            X, Y = self.example_to_text(examples)
            X = self.X_text_to_vec(X)
            Y = self.Y_text_to_vec(Y)
            testXs.extend(X)
            testYs.extend(Y)
        test_data = {
            'x': testXs,
            'y': testYs
        }
        with open(os.path.join(self.root, 'raw_data', 'test_data.json'), 'w') as f:
            json.dump(test_data, f)

    def _split_into_plays(self, shakespeare_full):
        """Splits the full data by play."""
        # List of tuples (play_name, dict from character to list of lines)
        plays = []
        discarded_lines = []  # Track discarded lines.
        slines = shakespeare_full.splitlines(True)[1:]
        # skip contents, the sonnets, and all's well that ends well
        author_count = 0
        start_i = 0
        for i, l in enumerate(slines):
            if 'by William Shakespeare' in l:
                author_count += 1
            if author_count == 1:
                start_i = i - 5
                break
        slines = slines[start_i:]
        current_character = None
        comedy_of_errors = False
        for i, line in enumerate(slines):
            # This marks the end of the plays in the file.
            if i > 124195 - start_i:
                break
            # This is a pretty good heuristic for detecting the start of a new play:
            if 'by William Shakespeare' in line:
                current_character = None
                characters = collections.defaultdict(list)
                # The title will be 2, 3, 4, 5, 6, or 7 lines above "by William Shakespeare".
                for j in range(2,8):
                    if slines[i - j].strip():
                        title = slines[i-j]
                        break
                title = title.strip()

                assert title, (
                        'Parsing error on line %d. Expecting title 2 or 3 lines above.' %
                        i)
                comedy_of_errors = (title == 'THE COMEDY OF ERRORS')
                # Degenerate plays are removed at the end of the algorithm.
                plays.append((title, characters))
                continue
            # match_character_regex
            match = (self.COE_CHARACTER_RE.match(line) if comedy_of_errors else self.CHARACTER_RE.match(line))
            if match:
                character, snippet = match.group(1), match.group(2)
                # Some character names are written with multiple casings, e.g., SIR_Toby
                # and SIR_TOBY. To normalize the character names, we uppercase each name.
                # Note that this was not done in the original preprocessing and is a
                # recent fix.
                character = character.upper()
                if not (comedy_of_errors and character.startswith('ACT ')):
                    characters[character].append(snippet)
                    current_character = character
                    continue
                else:
                    current_character = None
                    continue
            elif current_character:
                # _match_continuation_regex
                match = (self.COE_CONT_RE.match(line) if comedy_of_errors else self.CONT_RE.match(line))
                if match:
                    if comedy_of_errors and match.group(1).startswith('<'):
                        current_character = None
                        continue
                    else:
                        characters[current_character].append(match.group(1))
                        continue
            # Didn't consume the line.
            line = line.strip()
            if line and i > 2646:
                # Before 2646 are the sonnets, which we expect to discard.
                discarded_lines.append('%d:%s' % (i, line))
        # Remove degenerate "plays".
        return [play for play in plays if len(play[1]) > 1], discarded_lines

    def play_and_character(self, play, character):
        return re.sub('\\W+', '_',(play + '_' + character).replace(' ', '_'))

    def _get_examples_by_character(self, plays):
        skipped_characters = 0
        all_examples = collections.defaultdict(list)
        def add_examples(example_dict, example_tuple_list):
            for play, character, sound_bite in example_tuple_list:
                example_dict[self.play_and_character(
                    play, character)].append(sound_bite)
        users_and_plays = {}
        for play, characters in plays:
            curr_characters = list(characters.keys())
            for c in curr_characters:
                users_and_plays[self.play_and_character(play, c)] = play
            for character, sound_bites in characters.items():
                examples = [(play, character, sound_bite) for sound_bite in sound_bites]
                if len(examples) < 2:
                    skipped_characters += 1
                    # Skip characters with fewer than 2 lines since we need at least one
                    # train and one test line.
                    continue
                add_examples(all_examples, examples)
        return users_and_plays, all_examples, skipped_characters

    def example_to_text(self, examples):
        text = ' '.join(examples)
        text = re.sub(r"   *", r' ', text)
        X = []
        Y = []
        for i in range(0, len(text) - self.SEQ_LENGTH, 1):
            seq_in = text[i:i + self.SEQ_LENGTH]
            seq_out = text[i + self.SEQ_LENGTH]
            X.append(seq_in)
            Y.append(seq_out)
        return X, Y

    def X_text_to_vec(self, X):
        return [[self.ALL_LETTERS.find(c) for c in word] for word in X]

    def Y_text_to_vec(self, Y):
        return [self.ALL_LETTERS.find(c) for c in Y]


class TaskGenerator(BasicTaskGenerator):
    def __init__(self, rawdata_path=os.path.join(flgo.benchmark.data_root, 'SHAKESPEARE')):
        super(TaskGenerator, self).__init__(benchmark='shakespeare_classification', rawdata_path=rawdata_path)
        # Regular expression to capture an actors name, and line continuation

    def load_data(self):
        self.train_data = SHAKESPEARE(self.rawdata_path, train=True)
        self.test_data = SHAKESPEARE(self.rawdata_path, train=False)
        return

    def partition(self):
        self.local_datas = self.partitioner(self.train_data)

class TaskPipe(BasicTaskPipe):
    class TaskDataset(torch.utils.data.Subset):
        def __init__(self, dataset, indices):
            super().__init__(dataset, indices)
            self.dataset = dataset
            self.indices = indices

        def __getitem__(self, idx):
            if isinstance(idx, list):
                return self.dataset[[self.indices[i] for i in idx]]
            return self.dataset[self.indices[idx]]

    def save_task(self, generator):
        client_names = self.gen_client_names(len(generator.local_datas))
        feddata = {'client_names': client_names, 'server_data': list(range(len(generator.test_data))),
                   'rawdata_path': generator.rawdata_path}
        for cid in range(len(client_names)): feddata[client_names[cid]] = {'data': generator.local_datas[cid], }
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)
        return

    def load_data(self, running_time_option) -> dict:
        # load the datasets
        train_data = SHAKESPEARE(root=self.feddata['rawdata_path'], train=True)
        test_data = SHAKESPEARE(root=self.feddata['rawdata_path'], train=False)
        # rearrange data for server
        server_data_test, server_data_val = self.split_dataset(test_data, running_time_option['test_holdout'])
        task_data = {'server': {'test': server_data_test, 'val': server_data_val}}
        # rearrange data for clients
        for cid, cname in enumerate(self.feddata['client_names']):
            cdata = self.TaskDataset(train_data, self.feddata[cname]['data'])
            if running_time_option['train_holdout'] > 0.0:
                cdata_train, cdata_val = self.split_dataset(cdata, running_time_option['train_holdout'])
                if running_time_option['local_test'] and cdata_val is not None and len(cdata_val)>1:
                    cdata_val, cdata_test = self.split_dataset(cdata_val, running_time_option['local_test_ratio'])
                else:
                    cdata_test = None
            else:
                cdata_train, cdata_val, cdata_test = cdata, None, None
            task_data[cname] = {'train': cdata_train, 'val': cdata_val, 'test': cdata_test}
        return task_data

class TaskCalculator(BasicTaskCalculator):
    def __init__(self, device, optimizer_name='sgd'):
        super(TaskCalculator, self).__init__(device, optimizer_name)
        self.criterion = torch.nn.CrossEntropyLoss()
        self.DataLoader = torch.utils.data.DataLoader

    def compute_loss(self, model, data):
        """
        Args: model: the model to train
                 data: the training dataset
        Returns: dict of train-one-step's result, which should at least contains the key 'loss'
        """
        tdata = self.to_device(data)
        outputs = model(tdata[0])
        loss = self.criterion(outputs, tdata[-1])
        return {'loss': loss}

    @torch.no_grad()
    def test(self, model, dataset, batch_size=64, num_workers=0, pin_memory=False):
        """
        Metric = [mean_accuracy, mean_loss]
        Args:
            dataset:
                 batch_size:
        Returns: [mean_accuracy, mean_loss]
        """
        model.eval()
        if batch_size==-1:batch_size=len(dataset)
        data_loader = self.get_dataloader(dataset, batch_size=batch_size, num_workers=num_workers, pin_memory=pin_memory)
        total_loss = 0.0
        num_correct = 0
        for batch_id, batch_data in enumerate(data_loader):
            batch_data = self.to_device(batch_data)
            outputs = model(batch_data[0])
            batch_mean_loss = self.criterion(outputs, batch_data[-1]).item()
            y_pred = outputs.data.max(1, keepdim=True)[1]
            correct = y_pred.eq(batch_data[-1].data.view_as(y_pred)).long().cpu().sum()
            num_correct += correct.item()
            total_loss += batch_mean_loss * len(batch_data[-1])
        return {'accuracy': 1.0*num_correct/len(dataset), 'loss':total_loss/len(dataset)}

    def to_device(self, data):
        return data[0].to(self.device), data[1].to(self.device)

    def get_dataloader(self, dataset, batch_size=64, shuffle=True, num_workers=0, pin_memory=False, drop_last=False, *args, **kwargs):
        if self.DataLoader == None:
            raise NotImplementedError("DataLoader Not Found.")
        return self.DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, pin_memory=pin_memory)


import torch
import torch.nn as nn
import torch.nn.parallel
import torch.utils.data
from torch.autograd import Variable
import numpy as np
import torch.nn.functional as F

class STN3d(nn.Module):
    def __init__(self, channel):
        super(STN3d, self).__init__()
        self.conv1 = torch.nn.Conv1d(channel, 64, 1)
        self.conv2 = torch.nn.Conv1d(64, 128, 1)
        self.conv3 = torch.nn.Conv1d(128, 1024, 1)
        self.fc1 = nn.Linear(1024, 512)
        self.fc2 = nn.Linear(512, 256)
        self.fc3 = nn.Linear(256, 9)
        self.relu = nn.ReLU()

        self.bn1 = nn.BatchNorm1d(64)
        self.bn2 = nn.BatchNorm1d(128)
        self.bn3 = nn.BatchNorm1d(1024)
        self.bn4 = nn.BatchNorm1d(512)
        self.bn5 = nn.BatchNorm1d(256)

    def forward(self, x):
        batchsize = x.size()[0]
        x = F.relu(self.bn1(self.conv1(x)))
        x = F.relu(self.bn2(self.conv2(x)))
        x = F.relu(self.bn3(self.conv3(x)))
        x = torch.max(x, 2, keepdim=True)[0]
        x = x.view(-1, 1024)

        x = F.relu(self.bn4(self.fc1(x)))
        x = F.relu(self.bn5(self.fc2(x)))
        x = self.fc3(x)

        iden = Variable(torch.from_numpy(np.array([1, 0, 0, 0, 1, 0, 0, 0, 1]).astype(np.float32))).view(1, 9).repeat(
            batchsize, 1)
        x = x + iden.to(x.device)
        x = x.view(-1, 3, 3)
        return x


class STNkd(nn.Module):
    def __init__(self, k=64):
        super(STNkd, self).__init__()
        self.conv1 = torch.nn.Conv1d(k, 64, 1)
        self.conv2 = torch.nn.Conv1d(64, 128, 1)
        self.conv3 = torch.nn.Conv1d(128, 1024, 1)
        self.fc1 = nn.Linear(1024, 512)
        self.fc2 = nn.Linear(512, 256)
        self.fc3 = nn.Linear(256, k * k)
        self.relu = nn.ReLU()

        self.bn1 = nn.BatchNorm1d(64)
        self.bn2 = nn.BatchNorm1d(128)
        self.bn3 = nn.BatchNorm1d(1024)
        self.bn4 = nn.BatchNorm1d(512)
        self.bn5 = nn.BatchNorm1d(256)

        self.k = k

    def forward(self, x):
        batchsize = x.size()[0]
        x = F.relu(self.bn1(self.conv1(x)))
        x = F.relu(self.bn2(self.conv2(x)))
        x = F.relu(self.bn3(self.conv3(x)))
        x = torch.max(x, 2, keepdim=True)[0]
        x = x.view(-1, 1024)

        x = F.relu(self.bn4(self.fc1(x)))
        x = F.relu(self.bn5(self.fc2(x)))
        x = self.fc3(x)

        iden = Variable(torch.from_numpy(np.eye(self.k).flatten().astype(np.float32))).view(1, self.k * self.k).repeat(
            batchsize, 1)
        x = x + iden.to(x.device)
        x = x.view(-1, self.k, self.k)
        return x

class PointNetEncoder(nn.Module):
    def __init__(self, global_feat=True, feature_transform=False, channel=3):
        super(PointNetEncoder, self).__init__()
        self.stn = STN3d(channel)
        self.conv1 = torch.nn.Conv1d(channel, 64, 1)
        self.conv2 = torch.nn.Conv1d(64, 128, 1)
        self.conv3 = torch.nn.Conv1d(128, 1024, 1)
        self.bn1 = nn.BatchNorm1d(64)
        self.bn2 = nn.BatchNorm1d(128)
        self.bn3 = nn.BatchNorm1d(1024)
        self.global_feat = global_feat
        self.feature_transform = feature_transform
        if self.feature_transform:
            self.fstn = STNkd(k=64)

    def forward(self, x):
        B, D, N = x.size()
        trans = self.stn(x)
        x = x.transpose(2, 1)
        if D > 3:
            feature = x[:, :, 3:]
            x = x[:, :, :3]
        x = torch.bmm(x, trans)
        if D > 3:
            x = torch.cat([x, feature], dim=2)
        x = x.transpose(2, 1)
        x = F.relu(self.bn1(self.conv1(x)))

        if self.feature_transform:
            trans_feat = self.fstn(x)
            x = x.transpose(2, 1)
            x = torch.bmm(x, trans_feat)
            x = x.transpose(2, 1)
        else:
            trans_feat = None

        pointfeat = x
        x = F.relu(self.bn2(self.conv2(x)))
        x = self.bn3(self.conv3(x))
        x = torch.max(x, 2, keepdim=True)[0]
        x = x.view(-1, 1024)
        if self.global_feat:
            return x, trans, trans_feat
        else:
            x = x.view(-1, 1024, 1).repeat(1, 1, N)
            return torch.cat([x, pointfeat], 1), trans, trans_feat

def feature_transform_reguliarzer(trans):
    d = trans.size()[1]
    I = torch.eye(d)[None, :, :]
    loss = torch.mean(torch.norm(torch.bmm(trans, trans.transpose(2, 1)) - I.to(trans.device), dim=(1, 2)))
    return loss




from typing import Any
from collections.abc import Callable
# from ...cv.classification import GeneralCalculator
from torch.utils.data import random_split, Subset
from torchtext.data.functional import to_map_style_dataset
from flgo.benchmark.base import BasicTaskCalculator, BasicTaskGenerator, BasicTaskPipe
import os
try:
    import ujson as json
except:
    import json
import torch

class DataPipeGenerator(BasicTaskGenerator):
    def __init__(self, benchmark:str, rawdata_path:str, build_datapipes:Callable):
        super(DataPipeGenerator, self).__init__(benchmark, rawdata_path)
        self.build_datapipes = build_datapipes
        self.additional_option = {}
        self.train_additional_option = {}
        self.test_additional_option = {}

    def load_data(self, *args, **kwargs):
        # load train datapipe and convert it to train dataset
        train_options = self.additional_option.copy()
        train_options.update(self.train_additional_option)
        train_dp = self.build_datapipes(**train_options)
        train_dp = train_dp.map(lambda x: {'feature': x[0], 'label': x[1]})
        train_dp = train_dp.add_index('index')
        train_dp = train_dp.map(lambda x: (x['index'], (x['feature'], x['label'])))
        self.train_data = train_dp.to_map_datapipe()

    def partition(self, *args, **kwargs):
        self.local_datas = self.partitioner(self.train_data)
        self.num_clients = len(self.local_datas)

class DataPipeTaskPipe(BasicTaskPipe):
    TaskDataset = Subset
    def __init__(self, task_path, build_datapipes):
        super(DataPipeTaskPipe, self).__init__(task_path)
        self.build_datapipes = build_datapipes

    def save_task(self, generator):
        client_names = self.gen_client_names(len(generator.local_datas))
        feddata = {'client_names': client_names, 'rawdata_path': generator.rawdata_path, 'additional_option': generator.additional_option, 'train_additional_option':generator.train_additional_option, 'test_additional_option':generator.test_additional_option, }
        for cid in range(len(client_names)): feddata[client_names[cid]] = {'data': generator.local_datas[cid],}
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)
        return

    def load_data(self, running_time_option) -> dict:
        def feat2dict(x):
            return {'Feat_'+str(k):v for k,v in enumerate(x)}

        def fdict2tuple(x):
            return (x['__index__'], tuple(x.values()))
        # load train datapipe and convert it to train dataset
        train_options = self.feddata['additional_option'].copy()
        train_options.update(self.feddata['train_additional_option'])
        train_dp = self.build_datapipes(**train_options)
        train_data = to_map_style_dataset(train_dp)
        # load test datapipe and convert it to test dataset
        test_options = self.feddata['additional_option'].copy()
        test_options.update(self.feddata['train_additional_option'])
        test_dp = self.build_datapipes(**test_options)
        test_data = to_map_style_dataset(test_dp)
        # rearrange data for server
        server_data_test, server_data_valid = self.split_dataset(test_data, running_time_option['test_holdout'])
        task_data = {'server': {'test': server_data_test, 'val': server_data_valid}}
        # rearrange data for clients
        for cid, cname in enumerate(self.feddata['client_names']):
            cdata = self.TaskDataset(train_data, self.feddata[cname]['data'])
            if running_time_option['train_holdout'] > 0:
                cdata_train, cdata_val = self.split_dataset(cdata, running_time_option['train_holdout'])
                if running_time_option['local_test']:
                    cdata_val, cdata_test = self.split_dataset(cdata_val, running_time_option['local_test_ratio'])
                else:
                    cdata_test = None
            else:
                cdata_train = cdata
                cdata_val, cdata_test = None, None
            task_data[cname] = {'train': cdata_train, 'val': cdata_val, 'test': cdata_test}
        return task_data

class GeneralCalculator(BasicTaskCalculator):
    r"""
    Calculator for the dataset in torchvision.datasets.

    Args:
        device (torch.device): device
        optimizer_name (str): the name of the optimizer
    """
    def __init__(self, device, optimizer_name='sgd'):
        super(GeneralCalculator, self).__init__(device, optimizer_name)
        self.criterion = torch.nn.CrossEntropyLoss()
        self.DataLoader = torch.utils.data.DataLoader

    def compute_loss(self, model, data):
        """
        Args:
            model: the model to train
            data: the training dataset
        Returns: dict of train-one-step's result, which should at least contains the key 'loss'
        """
        text, label  = self.to_device(data)
        outputs = model(text)
        loss = self.criterion(outputs, label)
        return {'loss': loss}

    @torch.no_grad()
    def test(self, model, dataset, batch_size=64, num_workers=0, pin_memory=False):
        """
        Metric = [mean_accuracy, mean_loss]

        Args:
            model:
            dataset:
            batch_size:
        Returns: [mean_accuracy, mean_loss]
        """
        model.eval()
        if batch_size==-1:batch_size=len(dataset)
        data_loader = self.get_dataloader(dataset, batch_size=batch_size, num_workers=num_workers, pin_memory=pin_memory)
        total_loss = 0.0
        num_correct = 0
        for batch_id, batch_data in enumerate(data_loader):
            batch_data = self.to_device(batch_data)
            outputs = model(batch_data[0])
            batch_mean_loss = self.criterion(outputs, batch_data[1]).item()
            # y_pred = outputs.data.max(1, keepdim=True)[1]
            # correct = y_pred.eq(batch_data[-1].data.view_as(y_pred)).long().cpu().sum()
            num_correct += (outputs.argmax(1)==batch_data[1]).sum().item()
            total_loss += batch_mean_loss * len(batch_data[0])
        return {'accuracy': 1.0*num_correct/len(dataset), 'loss':total_loss/len(dataset)}

    def to_device(self, data):
        res = []
        for i in range(len(data)):
            if isinstance(data[i], torch.Tensor):
                di = data[i].to(self.device)
            elif isinstance(data[i], list):
                di = [d.to(self.device) for d in data[i]]
            else:
                raise TypeError('data should be either of type list or torch.Tensor')
            res.append(di)
        return tuple(res)

    def get_dataloader(self, dataset, batch_size=64, shuffle=True, num_workers=0, pin_memory=False, drop_last=False, *args, **kwargs):
        if self.DataLoader == None:
            raise NotImplementedError("DataLoader Not Found.")
        return self.DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, pin_memory=pin_memory, drop_last=drop_last, collate_fn=self.collate_fn)



import math

import flgo
import numpy
import torch
import os
import flgo.benchmark.base as fbb
from torch.utils.data import Subset
import torch.nn.functional as F
try:
    import ujson as json
except:
    import json
class BuiltinClassGenerator(fbb.BasicTaskGenerator):
    r"""
    Generator for the time series dataset.

    Args:
        benchmark (str): the name of the benchmark
        rawdata_path (str): the path storing the raw data
        builtin_class (class): dataset class
    """
    def __init__(self, benchmark, rawdata_path, builtin_class):
        super(BuiltinClassGenerator, self).__init__(benchmark, rawdata_path)
        self.builtin_class = builtin_class
        self.additional_option = {}
        self.train_additional_option = {}
        self.test_additional_option = {}

    def load_data(self):
        # load the datasets
        train_default_init_para = {'root': self.rawdata_path, 'train': True}
        test_default_init_para = {'root': self.rawdata_path, 'train': False}
        train_default_init_para.update(self.additional_option)
        train_default_init_para.update(self.train_additional_option)
        test_default_init_para.update(self.additional_option)
        test_default_init_para.update(self.test_additional_option)
        # if 'kwargs' not in self.builtin_class.__init__.__annotations__:
        #     train_pop_key = [k for k in train_default_init_para.keys() if
        #                      k not in self.builtin_class.__init__.__annotations__]
        #     test_pop_key = [k for k in test_default_init_para.keys() if
        #                     k not in self.builtin_class.__init__.__annotations__]
        #     for k in train_pop_key: train_default_init_para.pop(k)
        #     for k in test_pop_key: test_default_init_para.pop(k)
        # init datasets
        self.train_data = self.builtin_class(**train_default_init_para)
        self.test_data = self.builtin_class(**test_default_init_para)

    def partition(self):
        self.local_datas = self.partitioner(self.train_data)
        self.num_clients = len(self.local_datas)

class BuiltinClassPipe(fbb.BasicTaskPipe):
    r"""
    TaskPipe for the time series dataset.

    Args:
        task_path (str): the path of the task
        builtin_class (class): dataset class
    """
    class TaskDataset(torch.utils.data.Subset):
        def __init__(self, dataset, indices, perturbation=None, pin_memory=False):
            super().__init__(dataset, indices)
            self.dataset = dataset
            self.indices = indices
            self.perturbation = {idx:p for idx, p in zip(indices, perturbation)} if perturbation is not None else None
            self.pin_memory = pin_memory
            if not self.pin_memory:
                self.X = None
                self.Y = None
            else:
                self.X = torch.stack([self.dataset[i][0] for i in self.indices])
                self.Y = torch.stack([self.dataset[i][1] for i in self.indices])

        def __getitem__(self, idx):
            if self.X is not None:
                if self.perturbation is None:
                    return self.X[idx], self.Y[idx]
                else:
                    return self.X[idx]+self.perturbation[self.indices[idx]], self.Y[idx]
            else:
                if self.perturbation is None:
                    if isinstance(idx, list):
                        return self.dataset[[self.indices[i] for i in idx]]
                    return self.dataset[self.indices[idx]]
                else:
                    return self.dataset[self.indices[idx]][0] + self.perturbation[self.indices[idx]],  self.dataset[self.indices[idx]][1]

    def __init__(self, task_path, buildin_class):
        """
        Args:
            task_path (str): the path of the task
            builtin_class (class): dataset class
        """
        super(BuiltinClassPipe, self).__init__(task_path)
        self.builtin_class = buildin_class

    def save_task(self, generator):
        client_names = self.gen_client_names(len(generator.local_datas))
        feddata = {'client_names': client_names, 'server_data': list(range(len(generator.test_data))),  'rawdata_path': generator.rawdata_path, 'additional_option': generator.additional_option, 'train_additional_option':generator.train_additional_option, 'test_additional_option':generator.test_additional_option,}
        for cid in range(len(client_names)): feddata[client_names[cid]] = {'data': generator.local_datas[cid],}
        if hasattr(generator.partitioner, 'local_perturbation'): feddata['local_perturbation'] = generator.partitioner.local_perturbation
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)
        return

    def load_data(self, running_time_option) -> dict:
        # load the datasets
        train_default_init_para = {'root': self.feddata['rawdata_path'], 'train': True}
        test_default_init_para = {'root': self.feddata['rawdata_path'], 'train': False}
        if 'additional_option' in self.feddata.keys():
            train_default_init_para.update(self.feddata['additional_option'])
            test_default_init_para.update(self.feddata['additional_option'])
        if 'train_additional_option' in self.feddata.keys(): train_default_init_para.update(self.feddata['train_additional_option'])
        if 'test_additional_option' in self.feddata.keys(): test_default_init_para.update(self.feddata['test_additional_option'])
        # if 'kwargs' not in self.builtin_class.__init__.__annotations__:
        #     train_pop_key = [k for k in train_default_init_para.keys() if k not in self.builtin_class.__init__.__annotations__]
        #     test_pop_key = [k for k in test_default_init_para.keys() if k not in self.builtin_class.__init__.__annotations__]
        #     for k in train_pop_key: train_default_init_para.pop(k)
        #     for k in test_pop_key: test_default_init_para.pop(k)
        train_data = self.builtin_class(**train_default_init_para)
        test_data = self.builtin_class(**test_default_init_para)
        test_data = self.TaskDataset(test_data, list(range(len(test_data))), None, running_time_option['pin_memory'])
        # rearrange data for server
        server_data_test, server_data_val = self.split_dataset(test_data, running_time_option['test_holdout'])
        task_data = {'server': {'test': server_data_test, 'val': server_data_val}}
        # rearrange data for clients
        local_perturbation = self.feddata['local_perturbation'] if 'local_perturbation' in self.feddata.keys() else [None for _ in self.feddata['client_names']]
        for cid, cname in enumerate(self.feddata['client_names']):
            cpert = None if  local_perturbation[cid] is None else [torch.tensor(t) for t in local_perturbation[cid]]
            cdata = self.TaskDataset(train_data, self.feddata[cname]['data'], cpert, running_time_option['pin_memory'])
            cdata_train, cdata_val = self.split_dataset(cdata, running_time_option['train_holdout'])
            if running_time_option['train_holdout']>0 and running_time_option['local_test']:
                cdata_val, cdata_test = self.split_dataset(cdata_val, running_time_option['local_test_ratio'])
            else:
                cdata_test = None
            task_data[cname] = {'train':cdata_train, 'val':cdata_val, 'test': cdata_test}
        return task_data

class FromDatasetGenerator(fbb.FromDatasetGenerator):
    def __init__(self, seq_len=None, train_data=None, val_data=None, test_data=None):
        super(FromDatasetGenerator, self).__init__(benchmark=os.path.split(os.path.dirname(__file__))[-1],
                                            train_data=train_data, val_data=val_data, test_data=test_data)
        self.seq_len = seq_len
        if self.seq_len is not None:
            if self.train_data is not None and hasattr(self.train_data, 'set_len'):
                self.train_data.set_len(*seq_len)
            if self.test_data is not None and hasattr(self.test_data, 'set_len'):
                self.test_data.set_len(*seq_len)
            if self.val_data is not None and hasattr(self.val_data, 'set_len'):
                self.val_data.set_len(*seq_len)

class FromDatasetPipe(fbb.FromDatasetPipe):
    TaskDataset = Subset
    def __init__(self, task_path, train_data=None, val_data=None, test_data=None):
        super(FromDatasetPipe, self).__init__(task_path, train_data=train_data, val_data=val_data, test_data=test_data)

    def save_task(self, generator):
        client_names = self.gen_client_names(len(generator.local_datas))
        feddata = {'client_names': client_names}
        for cid in range(len(client_names)): feddata[client_names[cid]] = {'data': generator.local_datas[cid],}
        if hasattr(generator, 'seq_len') and generator.seq_len is not None:
            feddata['seq_len'] = generator.seq_len
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)
        return

    def split_dataset(self, dataset, p=0.0):
        if p == 0: return dataset, None
        s1 = int(len(dataset) * p)
        s2 = len(dataset) - s1
        if s1==0:
            return dataset, None
        elif s2==0:
            return None, dataset
        else:
            all_idx = list(range(len(dataset)))
            d1_idx = all_idx[:s2]
            d2_idx = all_idx[s2:]
            return Subset(dataset, d1_idx), Subset(dataset, d2_idx)

    def load_data(self, running_time_option) -> dict:
        train_data = self.train_data
        test_data = self.test_data
        val_data = self.val_data
        seq_len = self.feddata.get('seq_len', None)
        if seq_len is not None:
            if hasattr(self.train_data, 'set_len'):
                self.train_data.set_len(*seq_len)
            if hasattr(self.val_data, 'set_len'):
                self.val_data.set_len(*seq_len)
            if hasattr(self.test_data, 'set_len'):
                self.test_data.set_len(*seq_len)
        # rearrange data for server
        if val_data is None:
            server_data_test, server_data_val = self.split_dataset(test_data, running_time_option['test_holdout'])
        else:
            server_data_test = test_data
            server_data_val = val_data
        task_data = {'server': {'test': server_data_test, 'val': server_data_val}}
        # rearrange data for clients
        for cid, cname in enumerate(self.feddata['client_names']):
            cdata = self.TaskDataset(train_data, self.feddata[cname]['data'])
            cdata_train, cdata_val = self.split_dataset(cdata, running_time_option['train_holdout'])
            if running_time_option['train_holdout']>0 and running_time_option['local_test']:
                cdata_val, cdata_test = self.split_dataset(cdata_val, running_time_option['local_test_ratio'])
            else:
                cdata_test = None
            task_data[cname] = {'train':cdata_train, 'val':cdata_val, 'test': cdata_test}
        return task_data


class GeneralCalculator(fbb.BasicTaskCalculator):
    def __init__(self, device, optimizer_name='sgd'):
        super(GeneralCalculator, self).__init__(device, optimizer_name)
        self.criterion = torch.nn.MSELoss()
        self.DataLoader = torch.utils.data.DataLoader

    def to_device(self, data):
        if len(data)==2:
            return data[0].to(self.device), data[1].to(self.device)
        elif len(data)==4:
            x,y = data[0].to(self.device), data[1].to(self.device)
            if isinstance(data[2], torch.Tensor):
                return x,y, data[2].to(self.device), data[3].to(self.device)
            else:
                return x,y, data[2],data[3]

    def compute_loss(self, model, data):
        data = self.to_device(data)
        if len(data)==2:
            x, y = data
            ypred = model(x)
        elif len(data)==4:
            x, y, xmark, ymark = data
            ypred = model(x, xmark, ymark)
        loss = self.criterion(ypred, y)
        return {'loss': loss}

    @torch.no_grad()
    def test(self, model, dataset, batch_size=64, num_workers=0, pin_memory=False):
        """
        Metric = [mean_accuracy, mean_loss]

        Args:
            model:
            dataset:
            batch_size:
        Returns: [mean_accuracy, mean_loss]
        """
        model.eval()
        if batch_size == -1: batch_size = len(dataset)
        data_loader = self.get_dataloader(dataset, batch_size=batch_size, num_workers=num_workers, pin_memory=pin_memory)
        mse = 0.0
        mae = 0.0
        for batch_id, batch_data in enumerate(data_loader):
            batch_data = self.to_device(batch_data)
            if len(batch_data) == 2:
                x, y = batch_data
                outputs = model(x)
            elif len(batch_data) == 4:
                x, y, xmark, ymark = batch_data
                outputs = model(x, xmark, ymark)
            batch_mse = self.criterion(outputs, y).item()
            batch_mae = F.l1_loss(outputs, y).item()
            mse += batch_mse * len(batch_data[-1])
            mae += batch_mae * len(batch_data[-1])
        return {'mse': mse / len(dataset), 'mae': mae / len(dataset)}





import os

import torch
from torchvision.datasets.utils import download_and_extract_archive
import flgo
import torchvision
from flgo.benchmark.toolkits.cv.segmentation import BuiltinClassGenerator, BuiltinClassPipe, GeneralCalculator
from ..toolkits.cv.segmentation.utils import get_transform
from ..toolkits.cv.segmentation.coco_utils import FilterAndRemapCocoCategories, ConvertCocoPolysToMask
from ..toolkits.cv.segmentation.transforms import Compose
from flgo.benchmark.toolkits.cv.segmentation.coco_utils import _coco_remove_images_without_annotations
from ..toolkits import download_from_url, extract_from_zip

CAT_LIST = [0, 5, 2, 16, 9, 44, 6, 3, 17, 62, 21, 67, 18, 19, 4, 1, 64, 20, 63, 7, 72]
PATHS = {
        "train": ("train2017", os.path.join("annotations", "instances_train2017.json")),
        "val": ("val2017", os.path.join("annotations", "instances_val2017.json")),
        # "train": ("val2017", os.path.join("annotations", "instances_val2017.json"))
    }
builtin_class = torchvision.datasets.CocoDetection
train_transforms = Compose([FilterAndRemapCocoCategories(CAT_LIST, remap=True), ConvertCocoPolysToMask(), get_transform(train=True)])
test_transforms = Compose([FilterAndRemapCocoCategories(CAT_LIST, remap=True), ConvertCocoPolysToMask(), get_transform(train=False)])
path = os.path.join(flgo.benchmark.data_root,  'COCO')

class TaskGenerator(BuiltinClassGenerator):
    def __init__(self, rawdata_path=path):
        super(TaskGenerator, self).__init__(benchmark=os.path.split(os.path.dirname(__file__))[-1], rawdata_path=rawdata_path,
                                            builtin_class=builtin_class, train_transform=train_transforms, test_transform=test_transforms, num_classes = 21)
        self.additional_option = {}
        self.train_image_path = os.path.join(self.rawdata_path, "train2017")
        self.train_annotation_path = os.path.join(self.rawdata_path,"annotations", "instances_train2017.json")
        self.val_image_path = os.path.join(self.rawdata_path, "val2017")
        self.val_annotation_path = os.path.join(self.rawdata_path,"annotations", "instances_val2017.json")
        self.train_additional_option = {'root':self.train_image_path, 'annFile':self.train_annotation_path}
        self.test_additional_option = {'root':self.val_image_path, 'annFile':self.val_annotation_path}
        self.num_classes = 21

    def load_data(self):
        # download the dataset if the dataset doesn't exist
        if not os.path.exists(self.train_image_path):
            self.download_data('train')
        if not os.path.exists(self.val_image_path):
            self.download_data('val')
        if not os.path.exists(os.path.join(self.rawdata_path, "annotations")):
            self.download_data('anno')
        # load the datasets
        train_default_init_para = {'root': self.rawdata_path, 'download':self.download_data, 'train':True, 'transforms':self.train_transform}
        test_default_init_para = {'root': self.rawdata_path, 'download':self.download_data, 'train':False, 'transforms':self.test_transform}
        train_default_init_para.update(self.additional_option)
        train_default_init_para.update(self.train_additional_option)
        test_default_init_para.update(self.additional_option)
        test_default_init_para.update(self.test_additional_option)
        train_pop_key = [k for k in train_default_init_para.keys() if k not in self.builtin_class.__init__.__annotations__]
        test_pop_key = [k for k in test_default_init_para.keys() if k not in self.builtin_class.__init__.__annotations__]
        for k in train_pop_key: train_default_init_para.pop(k)
        for k in test_pop_key: test_default_init_para.pop(k)
        # init datasets
        self.train_data = self.builtin_class(**train_default_init_para)
        self.test_data = self.builtin_class(**test_default_init_para)
        self.train_data = _coco_remove_images_without_annotations(self.train_data, CAT_LIST)

    def download_data(self, flag:str):
        if flag=='train':
            url = "http://images.cocodataset.org/zips/train2017.zip"
            filename = "train2017.zip"
        elif flag=='val':
            url = "http://images.cocodataset.org/zips/val2017.zip"
            filename = "val2017.zip"
        elif flag=='anno':
            url = "http://images.cocodataset.org/annotations/annotations_trainval2017.zip"
            filename = "annotations_trainval2017.zip"
        download_and_extract_archive(url, download_root=self.rawdata_path, filename=filename)
        return

class TaskPipe(BuiltinClassPipe):
    def __init__(self, task_path):
        super(TaskPipe, self).__init__(task_path, builtin_class, train_transforms, test_transforms)

    def load_data(self, running_time_option) -> dict:
        # load the datasets
        train_default_init_para = {'root': self.feddata['rawdata_path'], 'download':True, 'train':True, 'transforms':self.train_transform}
        test_default_init_para = {'root': self.feddata['rawdata_path'], 'download':True, 'train':False, 'transforms':self.test_transform}
        if 'additional_option' in self.feddata.keys():
            train_default_init_para.update(self.feddata['additional_option'])
            test_default_init_para.update(self.feddata['additional_option'])
        if 'train_additional_option' in self.feddata.keys(): train_default_init_para.update(self.feddata['train_additional_option'])
        if 'test_additional_option' in self.feddata.keys(): test_default_init_para.update(self.feddata['test_additional_option'])
        train_pop_key = [k for k in train_default_init_para.keys() if k not in self.builtin_class.__init__.__annotations__]
        test_pop_key = [k for k in test_default_init_para.keys() if k not in self.builtin_class.__init__.__annotations__]
        for k in train_pop_key: train_default_init_para.pop(k)
        for k in test_pop_key: test_default_init_para.pop(k)
        train_data = self.builtin_class(**train_default_init_para)
        train_data = _coco_remove_images_without_annotations(train_data, CAT_LIST)
        test_data = self.builtin_class(**test_default_init_para)
        test_data = self.TaskDataset(test_data, list(range(len(test_data))), None, running_time_option['pin_memory'])
        # rearrange data for server
        server_data_test, server_data_val = self.split_dataset(test_data, running_time_option['test_holdout'])
        num_classes = self.feddata['num_classes']
        if server_data_val is not None: server_data_val.num_classes = num_classes
        if server_data_test is not None: server_data_test.num_classes = num_classes
        task_data = {'server': {'test': server_data_test, 'val': server_data_val}}
        # rearrange data for clients
        local_perturbation = self.feddata['local_perturbation'] if 'local_perturbation' in self.feddata.keys() else [None for _ in self.feddata['client_names']]
        for cid, cname in enumerate(self.feddata['client_names']):
            cpert = None if  local_perturbation[cid] is None else [torch.tensor(t) for t in local_perturbation[cid]]
            cdata = self.TaskDataset(train_data, self.feddata[cname]['data'], cpert, running_time_option['pin_memory'])
            cdata_train, cdata_val = self.split_dataset(cdata, running_time_option['train_holdout'])
            if running_time_option['train_holdout']>0 and running_time_option['local_test']:
                cdata_val, cdata_test = self.split_dataset(cdata_val, running_time_option['local_test_ratio'])
            else:
                cdata_test = None
            if cdata_train is not None: cdata_train.num_classes = num_classes
            if cdata_val is not None: cdata_val.num_classes = num_classes
            if cdata_test is not None: cdata_test.num_classes = num_classes
            task_data[cname] = {'train':cdata_train, 'val':cdata_val, 'test': cdata_test}
        return task_data

TaskCalculator = GeneralCalculator